# -*- coding: utf-8 -*-
"""
Feature class for the time series of profiles observation pattern.
"""
from __future__ import absolute_import, division, print_function

from typing import Tuple

from cerbere.dataset.dataset import STANDARD_Z_FIELD
from cerbere.feature.feature import Feature


class TimeSeriesProfile(Feature):
    """
    Feature class for the time series of profiles observation pattern.

    A :class:`~cerbere.feature.timeseriesprofile.TimeSeriesProfile` feature is
    defined by geolocation coordinates field, , all having a single geolocation
    dimension: ``profile`` (except ``z`` for Z axis coordinate).
      * ``lon(profile)``
      * ``lat(profile)``
      * ``time(profile)``
      * ``depth/alt (profile, z)``
    """
    _feature_geodimnames = 'profile', 'z',

    def __init__(self, *args, **kwargs):
        super(TimeSeriesProfile, self).__init__(*args, **kwargs)

        # CF specific attrs and vars for profile feature
        self._std_dataset.attrs['featureType'] = 'timeSeriesProfile'

        # mandatory coordinates
        for _, v in self._std_dataset.data_vars.items():
            v.attrs['coordinates'] = 'time lat lon z'

    def get_geocoord_dimnames(
            self, fieldname: str, *args, **kwargs) -> Tuple[str, ...]:
        if fieldname in STANDARD_Z_FIELD or self.coords[fieldname].is_axis('Z'):
            return 'profile', 'z',
        else:
            return 'profile',


class UniZTimeSeriesProfile(Feature):
    """
    Feature class for the time series of profiles observation pattern, where z
    is unique and identical for each profile in the series.

    A :class:`~cerbere.feature.timeseriesprofile.TimeSeriesProfile` feature is
    defined by geolocation coordinates field, , all having a single geolocation
    dimension: ``profile`` (except ``z`` for Z axis coordinate).
      * ``lon(profile)``
      * ``lat(profile)``
      * ``time(profile)``
      * ``depth/alt (z)``
    """
    def get_geocoord_dimnames(
            self, fieldname: str, *args, **kwargs) -> Tuple[str, ...]:
        if fieldname in STANDARD_Z_FIELD or self.coords[fieldname].is_axis('Z'):
            return 'z',
        else:
            return 'profile',



"""
Reader class for KNMI / OSI SAF scatterometer Level 2 NETCDF files

Example of file pattern:
ascat_20230210_165700_metopc_22116_eps_o_coa_3301_ovw.l2.nc
"""
from pathlib import Path
import re
import typing as T

from .basereader import BaseReader


KNMI_DIMS = {
    'numrows': 'row',
    'NUMROWS': 'row',
    'numcells': 'cell',
    'NUMCELLS': 'cell',
    'NUMAMBIGS': 'solutions',
    'NUMVIEWS': 'views'}


KNMI_VARS = {
    'row_time': 'time',
    'wvc_lat': 'lat',
    'wvc_lon': 'lon'
}


class OSISAFSCATL2(BaseReader):

    pattern = re.compile(r"[a-z]*_[0-9]{8}_[0-9]{6}.*_ovw.l2.nc$")
    engine: str = "netcdf4"
    description: str = "Use OSI SAF NetCDF scatterometer files in Xarray and " \
                   "cerbere"
    url: str = "https://link_to/your_backend/documentation"

    @classmethod
    def postprocess(cls, ds, **kwargs):

        ds = ds.rename_dims(
            {k: v for k, v in KNMI_DIMS.items() if k in ds.dims})
        ds = ds.swap_dims(
            {k: v for k, v in KNMI_VARS.items() if k in ds.variables})

        # normalize longitude to -180/180
        ds['lon'] = ds['lon'].where(ds['lon'] < 180, ds['lon'] - 360)

        return ds



# -*- coding: utf-8 -*-
"""
Class for the swath feature
"""
import typing as T
import logging
import warnings

import xarray as xr

import cerbere
from cerbere.feature.cbasefeature import BaseFeature
from cerbere.feature.cgrid import Grid


__all__ = ['Image']


class Image(BaseFeature):
    """
    Feature class for representing a image, a two-dimensional irregular grid
    along a satellite track, with one single time value associated.

    An image feature typically corresponds to a satellite "snapshot" like
    provided typically with a SAR (though technically it is more a swath) or a
    high resolution optical imager: a spatially limited swath section, so small
    that we don't need a time value for each pixel or even scan line. A single
    time value is enough to describe it, as if it was an instant snapshot.
    """
    cdm_data_type = 'curvilinear'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'longitude': ('X', 'Y',),
        'latitude': ('X', 'Y',),
        'time': ['()', ('T',)],
        'vertical': ('Z',),
        'data': ('X', 'Y',)
    }

    # optional axes for the feature
    cf_optional_axes = ['vertical']

    # axis naming
    cb_axis_naming = {'X': 'across_track', 'Y': 'along-track'}

    def cerberize(
            cls,
            dataset: xr.Dataset,
            force_reorder: bool = True,
            instance_class: 'BaseFeature' = None
    ) -> xr.Dataset:

        # squeeze time dimension in non-coordinate fields
        # ctime = self.dataset.time.copy()
        # if 'time' in self.dataset.dims:
        #     self.dataset = self.dataset.squeeze(dim='time')
        # self.dataset.coords['time'] = ctime

        return super().cerberize(dataset, force_reorder, instance_class)

    def check_feature(cls, dataset: xr.Dataset) -> bool:
        with warnings.catch_warnings(category=cerbere.CWarning):
            warnings.simplefilter("ignore")
            return Grid.check_feature(dataset)

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[BaseFeature]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if cls._unexpected_cdm_data_type(cfdst):
            return

        # check horizontal axes are defined or at least some mapping
        if cfdst.cb.X is not None or cfdst.cb.Y is not None:
            return

        if cfdst.cb.cf_axis_dims['X'] == cfdst.cb.cf_axis_dims['Y']:
            return

        # check lat/lon/time are two-dimensional and share the same
        # dimensions
        if set(cfdst['lon'].dims) != set(cfdst['lat'].dims):
            return

        # check time is one-dimensional and has only one single value
        if cfdst['time'].sizes != (1,) and cfdst['time'].dims[0] != 'time':
            return

        if cfdst.cb.vertical is not None and set(cfdst['lon'].dims) != set(
                cfdst.cb.vertical.dims):
            return
        if len(set(cfdst['lon'].dims)) != 2:
            return

        return Image



from pathlib import Path
import re
import typing as T

import xarray as xr


class BaseReader:
    """Base class for cerbere readers"""

    pattern: str = None
    engine: str = "netcdf4"
    description: str = "Use <your dataset> files in Xarray"
    url: str = "https://link_to/your_contrib/documentation"

    @classmethod
    def open_dataset(cls, filename_or_obj: T.Union[str, Path], **kwargs):
        ds = xr.open_dataset(filename_or_obj, **kwargs)
        ds.encoding['source'] = filename_or_obj
        ds.encoding['reader'] = cls.__class__.__name__

        return ds

    @classmethod
    def open_mfdataset(
            cls, filename_or_obj: T.List[T.Union[str, Path]], **kwargs):
        ds = xr.open_mfdataset(filename_or_obj, **kwargs)
        ds.encoding['source'] = filename_or_obj
        ds.encoding['reader'] = cls.__class__.__name__

        return ds

    @classmethod
    def guess_can_open(cls, filename_or_obj: T.Union[str, Path]) -> bool:
        return re.match(cls.pattern, Path(filename_or_obj).name) is not None

    @classmethod
    def postprocess(cls, ds, decode_times: bool = True):
        return ds

    @classmethod
    def preprocess(cls, ds: xr.Dataset, **kwargs):
        """called before writing the data

        Put here all transformations required for the data to match given
        format specification associated with the reader
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Feature class for incomplete multidimensional collection of features.
"""
import typing as T

import xarray as xr

import cerbere
from cerbere.feature.cbasecollection import BaseCollection
from cerbere.feature.cbasefeature import BaseFeature

__all__ = ['IMDCollection']


class IMDCollection(BaseCollection):
    """
    Feature class for an incomplete multidimensional collection of features,
    as defined by Climate and Forecast convention:
    https://cfconventions.org/cf-conventions/cf-conventions.html
    #_incomplete_multidimensional_array_representation

    A :class:`~cerbere.feature.imdcollection.IMDCollection` class is defined by
    geolocation coordinates field and geolocation dimension of the feature, plus
    the instance dimension defined for this feature.

    For instance, for a :class:`~cerbere.feature.profile.Profile` feature, the
    instance dimension is ``profile``; the CF convention states that the
    geolocation coordinates field should be defined as: ``lon(profile)``,
    ``lat(profile)``, ``time(profile)``, ``z(profile, z)``; observation field
    should be defined as ``fieldname(profile, z)``.
    """

    @classmethod
    def from_list(
            cls,
            items: T.List[BaseFeature],
            feature_class: str = None) -> 'IMDCollection':
        """Build a collection feature from a list of features"""
        if feature_class is None:
            feature_class = items[0].__class__

        dst = xr.concat(
            [_.dataset for _ in items], dim=items[0]._instance_dimname
        ).rename_dims(
            {items[0]._instance_dimname: feature_class._instance_dimname})

        return IMDCollection(
            dst,
            feature_class=feature_class
        )

    @classmethod
    def guess_instance_var(cls, dataset: xr.Dataset):
        pass

    @classmethod
    def guess_instance_axis(cls, dataset: xr.Dataset):
        return dataset.cb.instance_dim

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset):

        for instance_class in ['Profile', 'Point']:
            try:
                cls._check_feature(dataset, instance_class=instance_class)
                return cls
            except cerbere.AxisError:
                pass

        # check if there is an instance variable
        instance_dims = dataset.cb.cf_instance_dims

        if len(instance_dims) == 0:
            return

        # to be a collection, the instance_var should have a dimension of
        # the same name
        for i_dim in instance_dims:
            if i_dim not in dataset.dims:
                return

        instance_dim = dataset.cb.cf_instance_dims[0]

        # verify conformity to IMD Collection
        # coordinates must be: X(i), Y(i), T(i), Z(i, z)
        inst_size = dataset.sizes[instance_dim]

        # lon(instance_dim,...)
        if not cls._check_coord_dims(dataset, 'lon', [instance_dim], inst_size):
            return
        # lat(instance_dim,...)
        if not cls._check_coord_dims(dataset, 'lat', [instance_dim], inst_size):
            return
        # time(instance_dim,...)
        if not cls._check_coord_dims(
                dataset, 'time', [instance_dim], inst_size):
            return

        # z(z)
        if dataset.cb.vertical is not None:
            z_size = dataset.cb.vertical.sizes
            if not len(z_size) == 2:
                return

        return cls



# encoding: utf-8
"""
Abstract class for all data feature objects
"""
from __future__ import absolute_import, division, print_function

import logging
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Mapping, Optional, Tuple

import xarray as xr

from ..dataset.dataset import Dataset
from ..dataset.field import Field

__all__ = ['Feature']


class Feature(Dataset, ABC):
    """
    Abstract class for all **cerbere** feature objects.

    .. |dataset| replace:: :mod:`~cerbere.dataset`
    .. |feature| replace:: :mod:`~cerbere.feature`

    All |feature| classes are inherited from this parent class.

    A feature object is a specialized |dataset| object as it
    can only contains typed observations, where the type is defined by the data
    structure or observation pattern. Being a |dataset| object,
    it can be created from the same arguments as any|dataset| class instance.

    Refer to |dataset|.

    In many cases, a |feature| object will be instantiated from a file content,
    read through a |dataset| class handling the specific format of this file.
    For instance:

    .. code-block:: python

        # let's read gridded data stored in NetCDF following GHRSST format
        # convention. Cerbere provides the GHRSSTNCDataset to read and normalize
        # this type of data.
        from cerbere.dataset.ghrsstncdataset import GHRSSTNCDataset

        ghrsstfile = GHRSSTNCDataset('my_ghrsst_file.nc')

        # map this content into a grid feature object
        gridfeature = Grid(ghrsstfile)

    Args:
        force_reorder (bool): force order of the feature dimensions to match the
            order defined for this feature (True by def.). Improves the level of
            standardization but will involve loading all data in memory in non
            Dask context.
    """
    def __init__(
            self,
            *args,
            force_reorder: Optional[bool]=True,
            **kwargs):
        """
        See:
        Dataset
        """
        super(Feature, self).__init__(*args, **kwargs)

        # check if the object contains coordinate information expected by the
        # current data model
        # check_model() is a virtual method to be implemented
        # by each derived data model class
        if not self._check_dimensions(force_reorder):
            raise ValueError('data structure does not match the feature type')

    @property
    def _dataset_class(self):
        if isinstance(self.dataset, Dataset):
            return self.dataset._dataset_class
        else:
            return 'Dataset'

    @property
    def _base_dataset(self):
        if isinstance(self.dataset, Dataset):
            return self.dataset
        else:
            return super(Feature, self)

    @property
    def _instance_dimname(self):
        raise NotImplementedError

    def __str__(self):
        result = 'Feature : %s\n' % self.__class__.__name__
        result += super(Feature, self).__str__()
        return result

    @property
    def feature_type(self) -> str:
        """Return the type of the feature"""
        return self.__class__.__name__

    @property
    @abstractmethod
    def _feature_geodimnames(self) -> Tuple[str, ...]:
        """The names of the geolocation dimensions defining the feature

        Returns:
            a tuple of geolocation dimension names, using their standard name.
        """
        raise NotImplementedError

    @property
    def geodims(self):
        """Returns the names and sizes of the geolocation dimensions.

        Returns:
            OrderedDict(str, int): a dict where keys as the names of the
                dimensions and the values their respective sizes.
        """
        return OrderedDict([
            (_, self.sizes[_]) for _ in list(self._feature_geodimnames)
            ])

    @abstractmethod
    def get_geocoord_dimnames(
            self, fieldname: str, values: 'xr.DataArray'
            ) -> Tuple[str, ...]:
        """Returns the CF compliant dimension names of a geolocation coordinate
        field.

        Args:
            fieldname (str): the name of the geolocation coordinate, among
                ``lat``, ``lon``, ``time``, ``depth`` or ``height``
            values (DataArray): data array of the coordinate field (used to
                discriminate features with ambiguous dimension sizes, such as
                grid which can have a single time value for all data or a time
                value defined for each pixel)

        Returns:
            tuple of str: the dimension names of the coordinate field

        Note:
            Abstract function to be overridden in each feature class.
        """
        raise NotImplementedError

    def _reorder_geodim(self):
        """check the spatial dimension order is consistent for each variable
        and reorder if necessary.

        Returns:
            tuple: lat and lon dimension names, as found in data vars
        """
        geodims = list(self._feature_geodimnames)
        for v in self._varnames:
            fielddims = list(self.get_field_dims(v))
            dims = [_ for _ in fielddims if _ in geodims]
            if len(dims) == len(geodims):
                if geodims != dims:
                    # re-order
                    logging.warning(
                        'Reordering some dimensions in {}'.format(v)
                    )
                    newdims = geodims + [
                        _ for _ in fielddims if _ not in geodims
                    ]
                    self._std_dataset[v] = self._std_dataset[v].transpose(
                        *newdims,
                        transpose_coords=True
                    )

    def _check_dimensions(self, force_reorder=True):
        """Check that the dataset loaded into the feature fits the required
        properties.

        Verify the existence of the required coordinate fields and dimensions.
        Enforce strict feature dimension order if force_reorder is True (by
        default).

        Args:
            force_reorder (bool): force order of the feature dimensions to match
                the order defined for this feature (True by def.). Improves the
                 level of standardization but will involve loading all data in
                 memory in non Dask context.
        """
        # check required time and space dimensions are in the dataset
        dim_validity = all([
            dim in self.dims
            for dim in self._feature_geodimnames
        ])

        if not dim_validity:
            for dim in self._feature_geodimnames:
                if dim not in self.dims:
                    logging.warning(
                        '  => Missing coordinate dimension: {}'.format(dim)
                    )

        # re-order the required dimensions in the expected order
        if force_reorder:
            self._reorder_geodim()

        return dim_validity

    def append(self,
               feature,
               prefix: str = '',
               add_coords: bool = False,
               as_new_dims: bool = False,
               fields: str = None):
        """Append the fields from another feature

        The fields can share the same coordinates and dimensions (if the added
        feature is an overlay of the current feature). In such case the
        shared coordinates of the added feature are not copied, unless
        `add_coords` is set to True (in which case they are first prefixed
        using `prefix`).

        If the dimensions of the added field have unrelated to the current
        feature (even if they have similar names), `as_new_dims` must be set
        to True. The dimensions (and corresponding coordinates) are prefixed
        with `prefix` and added to with the fields to the current feature.

        Args:
            feature (AbstractFeature derived class): the feature to append.

            prefix (str): prefix to use for naming the fields of the appended
                feature (to avoid conflicts with the existing fields of the
                current feature).

            add_coords: if True, add the feature coordinates as variables. If
                false, they are not added.

            as_new_dims: rename the field and coordinate dimensions with prefix

            fields (list of str): a list of fieldnames specifying which fields
                are to be appended. By default, all fields of the feature are
                appended.
        """
        if as_new_dims and prefix is None:
            raise ValueError(
                'a valid prefix must be provided to rename the dimensions')

        if add_coords and prefix is None:
            raise ValueError(
                'a valid prefix must be provided to add the coordinates')

        added_fields = fields
        if added_fields is None:
            added_fields = feature.fieldnames
        if add_coords:
            added_fields.extend(feature.coordnames)
        added_fields = list(set(added_fields))

        # append fields
        for fieldname in added_fields:
            if fieldname in feature.fieldnames:
                field = feature.get_field(fieldname).clone()
            elif fieldname in feature.coordnames:
                field = feature.get_coord(fieldname).clone()
            else:
                raise ValueError('field {} not found in feature'.format(
                    fieldname))
            field.rename('{}{}'.format(prefix, field.name))

            if as_new_dims:
                # we ensure dim and field names can not be the same (raise an
                # error with xarray
                field._array = field._array.swap_dims(
                    {_: '{}{}'.format(prefix, _) for _ in field.dims})

            self.add_field(field)

    def get_values(self, *args, **kwargs):
        """
        See:
         :meth:`~cerbere.dataset.Dataset.get_values`
        """
        #if 'expand_dims' not in kwargs:
        #    kwargs['expand_dims'] = self._feature_geodimnames
        return super(Feature, self).get_values(
            *args, **kwargs
        )

    def extract(
            self,
            *args,
            bbox=None,
            footprint=None,
            **kwargs) -> 'Feature':
        """Extract a subset as a new ``Feature`` object.

        The subset can be selected through one of the following:
        * bbox: a rectangular lat/lon bounding box.
        * footprint: a more complex shape provided as a lat/lon polygon.
        * index: any valid xarray selection index.

        If none of these are provided, a copy of the whole object is returned.
        The created subset is a new datamodel object of the same class without
        any reference to the source mapper.

        See:
             :func:`~cerbere.dataset.dataset.Dataset.extract`
        """
        if bbox is not None or footprint is not None:
            raise NotImplementedError
        return self.__class__(
            super(Feature, self).extract(*args, **kwargs))

    def extract_field(
            self,
            fieldname: str,
            index: Mapping[str, slice] = None,
            padding: bool = False,
            prefix: str = None,
            **kwargs):
        """
        Create a copy of a field, or subset of a field, and
        padding out as required.

        Args
            fieldname (str): The name of the field to extract.
            index (xarray index type, optional): any kind of xarray indexing

            padding (bool, optional): True to pad out feature with fill values
                to the extent of the dimensions.

            prefix (str, optional): add a prefix string to the field names of
               the extracted subset.
        """
        subset = self._get_xarray_index(index, **kwargs)

        if fieldname in self.coords:
            field = self.coords[fieldname]
        elif fieldname in self.data:
            field = self.data[fieldname]
        else:
            raise ValueError('Unknown field {}'.format(fieldname))

        new_field = Field.to_field(field).clone(subset, padding=padding)
        if prefix is not None:
            new_field.set_name(prefix + field.name)
        return new_field

#         new_dims = copy.copy(field.dimensions)
#         if slices:
#             for dim in new_dims.keys():
#                 if dim in slices:
#                     if not slices[dim].start:
#                         start = 0
#                     else:
#                         start = slices[dim].start
#                     if not slices[dim].stop:
#                         stop = field.dimensions[dim]
#                     else:
#                         stop = slices[dim].stop
#                     new_dims[dim] = stop - start
#
#         elif indices:
#             for dim in new_dims:
#                 if dim in indices:
#                     new_dims[dim] = len(indices[dim])
#
#         new_field = Field(
#             copy.copy(field.variable),
#             new_dims,
#             fields=copy.copy(field.components),
#             datatype=field.datatype
#         )
#         new_field.set_metadata(field.get_metadata())
#         new_field.set_values(
#             field.get_values(slices=slices, indices=indices, padding=padding)
#         )
#         if prefix is not None:
#             new_field.variable.shortname = (prefix +
#                                             new_field.variable.shortname)
#         return new_field



# -*- coding: utf-8 -*-
"""
Feature class for a set of randomly sampled points.
"""
from __future__ import absolute_import, division, print_function

from cerbere.feature.feature import Feature

__all__ = ['Point']


class Point(Feature):
    """
    Feature class for a set of randomly sampled points.

    A :class:`~cerbere.feature.point.Point` feature is defined in CF convention
    by geolocation coordinates field: ``lon(obs)``, ``lat(obs)``,
    ``time(obs)``, optionally ``z(obs)``, all having a single geolocation
    dimension: ``obs``.
    """
    _instance_dimname = 'obs'
    _feature_geodimnames = ()

    def __init__(self, *args, **kwargs):
        super(Point, self).__init__(*args, **kwargs)
        self._std_dataset.attrs['featureType'] = 'point'

    def get_geocoord_dimnames(self,  *args, **kwargs):
        return ()



# -*- coding: utf-8 -*-
"""
:mod:`~cerbere.dataset` class to read netCDF files using CF conventions.
"""
from __future__ import absolute_import, division, print_function

import datetime
import logging

from .dataset import Dataset

LOGGER = logging.getLogger()

NETCDF4 = 'NETCDF4'


class NCDataset(Dataset):
    """CF compliant NetCDF Dataset

    Args:
        format ({‘NETCDF4’, ‘NETCDF4_CLASSIC’, ‘NETCDF3_64BIT’,
            ‘NETCDF3_CLASSIC’}, optional): format (default is 'NETCDF4')

    See:
    :class:`~cerbere.dataset.dataset.Dataset`
    """
    def __init__(
            self,
            *args,
            format=NETCDF4,
            **kwargs):
        super().__init__(
            *args,
            format=format,
            **kwargs
        )


    def _convert_format(
        self,
        profile='ghrsst_saving_profile.yaml',
        **kwargs
    ):
        dataset = super()._convert_format(profile, **kwargs)

        # fill in some attributes
        creation_date = datetime.datetime.now()
        dataset.attrs['date_created'] = creation_date
        dataset.attrs['date_modified'] = creation_date
        return dataset

    def get_collection_id(self) -> str:
        """return the identifier of the product collection"""
        if 'id' in self.attrs:
            return self.attrs['id']



# encoding: utf-8
"""
Abstract class for all data feature objects
"""
from __future__ import absolute_import, division, print_function

import logging
import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict
import typing as T

import xarray as xr

import cerbere.accessor.cdataset


__all__ = ['BaseFeature']


class BaseFeature(object):
    """
    Abstract class for all **cerbere** feature objects.

    All |feature| classes are inherited from this parent class.

    A feature object contains a typed xarray |dataset| object
    corresponding to a specific observation pattern. It relies on,
    and sometimes complement, the patterns defined in Climate and Forecast (
    CF) convention. It harmonizes the naming conventions for spatial and
    temporal coordinates, dimensions, etc... and defines methods that can be
    specific to a specific type of feature.

    can only contain typed observations, where the type is defined by the data

    In many cases, a |feature| object will be instantiated from a file content,
    read through the cerbere `open_dataset` method. For instance:

    .. code-block:: python

        # let's read gridded data stored in NetCDF following GHRSST format
        # convention. Cerbere provides a specific engine to read and normalize
        # this type of data.
        import cerbere

        ghrsst_dst = cerbere.open_dataset('my_ghrsst_file.nc', engine='ghrsst')

        # map this content into a grid feature object
        gridfeature = cerbere.feature.Grid(ghrsst_dst)

    Args:
        force_reorder (bool): force order of the feature dimensions to match the
            order defined for this feature (True by def.). Improves the level of
            harmonization but will involve loading all data in memory in non
            Dask context.
    """

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'latitude': None, 'longitude': None, 'vertical': None, 'time': None,
        'data': None
    }
    # optional axes for the feature
    cf_optional_axes = []

    # Unidata Common Data Model provides a longer set of feature types
    cdm_data_type = None

    def __init__(
            self,
            dataset: xr.Dataset,
            force_reorder: bool = True,
            **kwargs):
        """
        See:
        Dataset
        """
        # ensure it is cerberized
        self.dataset = self.cerberize(dataset, force_reorder, **kwargs)

    @classmethod
    def cerberize(
            cls,
            dataset: xr.Dataset,
            force_reorder: bool = True,
            instance_class: 'BaseFeature' = None
    ) -> xr.Dataset:
        """check and modify a dataset so that it fully fits the
        cerbere/CF feature's data model
        """
        # check if the object contains coordinate information expected by the
        # current data model
        # check_model() is a virtual method to be implemented
        # by each derived data model class
        ds = cls._check_dimensions(
            dataset.cb.cfdataset,
            force_reorder=force_reorder,
            instance_class=instance_class)

        # fill in feature type attributes
        if cls.cdm_data_type is not None:
            ds.attrs['cdm_data_type'] = cls.cdm_data_type

        return ds

    @property
    def _instance_dimname(self):
        raise NotImplementedError

    def __str__(self):
        result = 'Feature : %s\n' % self.__class__.__name__
        result += super(BaseFeature, self).__str__()
        return result

    @property
    def ds(self) -> xr.Dataset:
        """The xarray Dataset object composing the feature"""
        return self.dataset

    @ds.setter
    def ds(self, dataset: xr.Dataset):
        self.dataset = dataset

    @property
    def type(self) -> str:
        """the cerbere class name of the feature"""
        return self.__class__.__name__

    @property
    def featureType(self) -> str:
        return self.dataset.attrs.get('featureType')

    @featureType.setter
    def featureType(self, value: str):
        self.dataset.attrs['featureType'] = value

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional['BaseFeature']:
        """Guess if the datasets fit the feature's data model.

        Returns:
            the class of the feature, if it's a match. None otherwise.
        """
        return

    # @classmethod
    # def required_dims(cls, *args, **kwargs) -> T.Tuple[str, ...]:
    #     """The names of the geolocation dimensions defining the feature class
    #
    #     Returns:
    #         a tuple of geolocation dimension names, using their standard name.
    #     """
    #     return cls._required_dims

    @property
    def geodims(self):
        """Returns the names and sizes of the geolocation dimensions.

        Returns:
            OrderedDict(str, int): a dict where keys as the names of the
                dimensions and the values their respective sizes.
        """
        return OrderedDict([
            (_, self.dataset.sizes[_]) for _ in list(self.cf_canonical_dims)
            ])

    @abstractmethod
    def get_geocoord_dimnames(
            self, fieldname: str, values: 'xr.DataArray'
            ) -> T.Tuple[str, ...]:
        """Returns the CF compliant dimension names of a geolocation coordinate
        field.

        Args:
            fieldname (str): the name of the geolocation coordinate, among
                ``lat``, ``lon``, ``time``, ``depth`` or ``height``
            values (DataArray): data array of the coordinate field (used to
                discriminate features with ambiguous dimension sizes, such as
                grid which can have a single time value for all data or a time
                value defined for each pixel)

        Returns:
            tuple of str: the dimension names of the coordinate field

        Note:
            Abstract function to be overridden in each feature class.
        """
        raise NotImplementedError

    @classmethod
    def _reorder_geodim(cls, dataset: xr.Dataset, instance_class=None) -> \
            xr.Dataset:
        """check the spatial dimension order is consistent for each variable
        and reorder if necessary.

        Returns:
            tuple: lat and lon dimension names, as found in data vars
        """
        geodims = list(cls.cf_canonical_dims)
        ds = dataset.copy(deep=False)
        for v in ds.data_vars:
            fielddims = list(ds[v].dims)
            dims = [_ for _ in fielddims if _ in geodims]
            if len(dims) == len(geodims):
                if geodims != dims:
                    # re-order
                    logging.warning(f'Reordering some dimensions in {v}')
                    newdims = geodims + [
                        _ for _ in fielddims if _ not in geodims
                    ]
                    ds[v] = ds[v].transpose(
                        *newdims,
                        transpose_coords=True
                    )

        return ds

    @classmethod
    def _unexpected_cdm_data_type(cls, cfdst):
        """check Unidata CDM' feature type"""
        return cfdst.cb.cdm_data_type is not None and \
            cls.cdm_data_type is not None and \
            cfdst.cb.cdm_data_type.lower() != cls.cdm_data_type.lower()

    @classmethod
    def _canonical_dims(
            cls, coord: str, *args, **kwargs
    ) -> T.Union[T.Set[str], T.List[T.Set[str]]]:
        return cls.cf_canonical_dims[coord]

    @classmethod
    def is_alias(cls, dataset):
        """Return the class of the parent the feature is an alias for

        Intended for special feature that are a particular implementation of
        an existing feature class (usually a collection), e.g.:
        TrajectoryProfile is an alias for a collection of profiles.
        """
        return

    @classmethod
    def _translate_cf_axis_dim(cls, cf_axis, dataset):
        """return the dimension name in dataset for a canonical CF dimension"""
        dim = dataset.cb.cf_axis_dims.get(cf_axis)
        if dim is not None:
            return dim

        if dataset.cb.cf_instance_dims is not None and len(
                dataset.cb.cf_instance_dims) > 0:
            return dataset.cb.cf_instance_dims[0]

        return cf_axis

    @classmethod
    def _check_feature(cls, dataset, instance_class=None):
        """Check that the dataset loaded into the feature fits the required
        properties.

        Verify the existence of the required coordinate fields and dimensions.
        Enforce strict feature dimension order if force_reorder is True (by
        default).

        Args:
            force_reorder (bool): force order of the feature dimensions to match
                the order defined for this feature (True by def.). Improves the
                 level of standardization but will involve loading all data in
                 memory in non Dask context.
            instance_class: only for collections
        """
        # required dimensions for dataset coordinates and variables
        if instance_class is not None:
            instance_class = cerbere.feature_class(instance_class)
            required_dims = instance_class.cf_canonical_dims
            cf_optional_coords = instance_class.cf_optional_axes
        else:
            required_dims = cls.cf_canonical_dims
            cf_optional_coords = cls.cf_optional_axes
        required_coords = set(required_dims.keys()) - {'data'}

        # check required spatiotemporal axis coordinates and dimensions are in
        # the dataset
        missing_coords = []
        incorrect_coords = []
        for cf_coord in required_coords:
            if cf_coord not in dataset.cb.cf_coords:
                if cf_coord not in set(cf_optional_coords):
                    missing_coords.append(cf_coord)
                continue

            coord = dataset.cb.cf_coords[cf_coord]

            # ds_canonical_dims = set([
            #     cls._translate_cf_axis_dim(_, dataset)
            #     for _ in cls._canonical_dims(cf_coord, instance_class)])
            #
            # if set(dataset[coord].dims) != ds_canonical_dims:
            #     incorrect_coords.append(coord)

            _c_dims = cls._canonical_dims(cf_coord, instance_class)
            if isinstance(_c_dims, list):
                ds_canonical_dims = [
                    set([cls._translate_cf_axis_dim(_, dataset)
                         for _ in alt]) for alt in _c_dims]
                if all([set(dataset[coord].dims) != _
                        for _ in ds_canonical_dims]):
                    incorrect_coords.append(coord)

            else:
                ds_canonical_dims = set([
                    cls._translate_cf_axis_dim(_, dataset) for _ in _c_dims])
                if set(dataset[coord].dims) != ds_canonical_dims and cf_coord\
                        not in cls.cf_optional_axes:
                    incorrect_coords.append(coord)

        if len(missing_coords) > 0:
            raise cerbere.AxisError(
                f'Data structure does not match the feature type. Missing '
                f'spatiotemporal coordinates: {missing_coords}')

        if len(incorrect_coords) > 0:
            log = ''
            for coord in incorrect_coords:
                log += (
                    f'Data structure does not match the feature type. Axis '
                    f'coordinate {coord} should have dimensions '
                    f'{cls.cf_canonical_dims[cf_coord]}\n')
            raise cerbere.AxisError(log)

    @classmethod
    def _check_dimensions(
            cls, dataset, force_reorder=True, instance_class=None):
        """Check that the dataset loaded into the feature fits the required
        properties.

        Verify the existence of the required coordinate fields and dimensions.
        Enforce strict feature dimension order if force_reorder is True (by
        default).

        Args:
            force_reorder (bool): force order of the feature dimensions to match
                the order defined for this feature (True by def.). Improves the
                 level of standardization but will involve loading all data in
                 memory in non Dask context.
            instance_class: only for collections
        """
        cls._check_feature(dataset, instance_class=instance_class)

        # # check required spatiotemporal axis coordinates and dimensions are in
        # # the dataset
        # missing_axis_coords = []
        # incorrect_axis_coords = []
        # for coord in required_dims:
        #     if coord not in dataset.cb.cf_axis_coords:
        #         missing_axis_coords.append(coord)
        #     ds_canonical_dims = set([
        #         dataset.cb.cf_axis_dims.get(_, _)
        #         for _ in cls._canonical_dims(coord)])
        #     if (set(dataset[dataset.cb.cf_axis_coords[coord]].dims) !=
        #             ds_canonical_dims):
        #         incorrect_axis_coords.append(coord)
        # if len(missing_axis_coords) > 0:
        #     raise cerbere.AxisError(
        #         f'Data structure does not match the feature type. Missing '
        #         f'axis coordinates: {missing_axis_coords}')
        # if len(incorrect_axis_coords) > 0:
        #     for coord in incorrect_axis_coords:
        #         logging.error(
        #             f'Data structure does not match the feature type. Axis '
        #             f'coordinate {coord} should have dimensions '
        #             f'{cls.cf_canonical_dims[coord]}')
        #     raise cerbere.AxisError(
        #         f'Data structure does not match the feature type.')

        # re-order the required dimensions in the expected canonical order
        if force_reorder:
            return cls._reorder_geodim(dataset, instance_class=instance_class)

        return dataset

    @classmethod
    def _check_dim_sizes(
            cls,
            arr_sizes: T.Dict,
            dims: T.Union[str, T.List[str]],
            sizes: T.Optional[T.Union[int, T.List[int]]] = None):
        if len(arr_sizes) == len(dims) == 0:
            return True

        if set(arr_sizes.keys()) != set(dims):
            return False

        if sizes is not None:
            if isinstance(dims, str):
                if not isinstance(sizes, int):
                    raise ValueError('sizes argument should be an integer')
                return arr_sizes[dims] == sizes

            if isinstance(sizes, int):
                sizes = [sizes] * len(dims)
            for i, dim in enumerate(dims):
                if arr_sizes[dim] != sizes[i]:
                    return False

        return True

    @classmethod
    def _check_coord_dims(
            cls,
            dataset: xr.Dataset,
            axis: str,
            dims: T.Union[str, T.List[str]],
            sizes: T.Optional[T.Union[int, T.List[int]]] = None) -> bool:
        """verify a dataset has a variable with the given dimensions and sizes.
        """
        dst = dataset.cb.cfdataset
        name = dataset.cb.cf_coords.get(axis)

        if name is None:
            return False

        return name in dst.coords and \
            set(dims) == set(dst.coords[name].dims) and \
            cls._check_dim_sizes(dst.coords[name].sizes, dims, sizes)

    def isel(self, *args, **kwargs) -> T.Any:
        """
        shortcut for cerbere ds.cb.isel method but returns the result
        wrapped into a Feature object of the same class as the current object
        """
        return self.__class__(self.ds.cb.isel(*args, **kwargs))

    def copy(self, deep: bool=False):
        return self.__class__(self.ds.copy(deep=deep))



# -*- coding: utf-8 -*-
"""
Feature class for the time series observation pattern at a fix station.
"""
import typing as T
import xarray as xr
import warnings

from cerbere.dataset.dataset import STANDARD_Z_FIELD
from cerbere import CWarning
from cerbere.feature.cdiscretefeature import DiscreteFeature

__all__ = ['TimeSeries']


class TimeSeries(DiscreteFeature):
    """
    This class is for a single time series. This feature is defined by
    geolocation coordinates field: ``lon()``, ``lat()``, ``z()``,
    ``time(time)``,  none having a geolocation dimension (except ``time`` for T
    axis coordinate).

    Refer to CF convention: https://cfconventions.org/cf-conventions/cf
    -conventions.html#time-series

    Data may be taken over periods of time at a set of discrete point,
    spatial locations called stations (see also discussion in 9.1).  The set
    of elements at a particular station is referred to as a timeSeries
    feature and a data variable may contain a collection of such features.
    The instance dimension in the case of timeSeries specifies the number of
    time series in the collection and is also referred to as the station
    dimension. The instance variables, which have just this dimension,
    including latitude and longitude for example, are also referred to as
    station variables and are considered to contain information describing
    the stations. The station variables may contain missing values, allowing
    one to reserve space for additional stations that may be added at a later
    time, as discussed in section 9.6. In addition,
    * It is strongly recommended that there should be a station variable (
      which may be of any type) with the attribute cf_role=”timeseries_id”,
      whose values uniquely identify the stations.
    * It is recommended that there should be station variables with
      standard_name attributes " platform_name ", " surface_altitude " and “
      platform_id ” when applicable.

    All the representations described in section 9.3 can be used for time
    series. The global attribute featureType=”timeSeries” (case-insensitive)
    must be included.
    """
    cf_instance_axis = 'station'
    cf_feature_type = 'timeSeries'
    cdm_data_type = 'station'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'latitude': (cf_instance_axis,),
        'longitude': (cf_instance_axis,),
        'time': (cf_instance_axis, 'T'),
        'vertical': (cf_instance_axis,),
        'data': (cf_instance_axis, 'T')
    }
    # optional axes for the feature
    cf_optional_axes = ['vertical']

    @classmethod
    def cerberize(cls, dataset: xr.Dataset, force_reorder: bool = True,
                  **kwargs):
        cfdataset = super(TimeSeries, cls).cerberize(dataset, force_reorder)

        # CF specific attrs and vars
        # It is strongly recommended that there should be a station variable
        # (which may be of any type) with the attribute
        # cf_role=”timeseries_id”, whose values uniquely identify the stations.
        # It is recommended that there should be station variables with
        # standard_name attributes " platform_name ", " surface_altitude "
        # and “ platform_id ” when applicable.
        if cfdataset.cb.cf_role is None:
            station_var = None
            for varname in cfdataset.variables:
                v = cfdataset.variables[varname]
                if v.attrs.get('standard_name') == 'platform_name' or \
                        v.attrs.get('standard_name') == 'platform_id':
                    cfdataset[v].cf_role = 'timeseries_id'
                    station_var = v
                    break

            if station_var is None:
                cfdataset['station_id'] = xr.DataArray(
                    dims=(),
                    attrs={'cf_role': 'timeseries_id'},
                    data=0
                )

        # for _, v in cfdataset.data_vars.items():
        #     v.encoding['coordinates'] = 'time lat lon z'

        return cfdataset

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[DiscreteFeature]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if cls._unexpected_featureType(cfdst) or \
                cls._unexpected_cdm_data_type(cfdst):
            return

        # check lat/lon are axes and single value
        for coordname in ['lon', 'lat']:
            if len(cfdst[coordname].dims) != 0 or cfdst[coordname].size != 1:
                return
        if set(cfdst['lon'].dims) != set(cfdst['lat'].dims):
            return

        # check time has only one axis which is time
        if len(cfdst['time'].dims) != 1 or 'time' not in cfdst['time'].dims:
            return

        # check provided axes if any
        if cfdst.cb.X is not None and cfdst.cb.X.name != 'lon':
            return
        if cfdst.cb.Y is not None and cfdst.cb.Y.name != 'lat':
            return
        if cfdst.cb.T is not None and cfdst.cb.T.name != 'time':
            return

        return TimeSeries



# -*- coding: utf-8 -*-
"""
Feature class for the profile observation pattern.
"""
import logging
import typing as T

import numpy as np
import xarray as xr

from cerbere.feature.cdiscretefeature import DiscreteFeature


__all__ = ['Profile']


class Profile(DiscreteFeature):
    """
    Feature class for the profile observation patterns.

    A :class:`~cerbere.feature.profile.Profile` feature is defined by
    coordinates:
    * ``lon()`` or ``lon(profile)``
    * ``lat()`` or ``lat(profile)``
    * ``time()`` or ``time(profile)``
    * ``z(z)``
    * ``profile()`` of any type

    """
    # must be defined for any feature corresponding to a CF discrete sampling
    # geometry, Nobe otherwise
    cf_feature_type = 'profile'

    # the instance dimension name for collection of features
    cf_instance_axis = 'profile'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    # x, y, z, t refer to the coordinates; X, Y, Z, T to the dimensions
    cf_canonical_dims = {
        'longitude': (cf_instance_axis,),
        'latitude': (cf_instance_axis,),
        'time': (cf_instance_axis,),
        'vertical': [(cf_instance_axis, 'Z'), ('Z',)],
        'data': (cf_instance_axis, 'Z')
    }
    # optional axes for the feature
    cf_optional_axes = []

    # Unidata Common Data Model provides a wider set of feature types than CF
    cdm_data_type = 'profile'

    @classmethod
    def cerberize(cls, dataset: xr.Dataset, force_reorder: bool = True,
                  **kwargs):
        """modify a dataset so that it fully fits the feature's data model"""
        cfdataset = super(Profile, cls).cerberize(dataset, force_reorder)

        # CF specific attrs and vars for profile feature
        if 'profile' not in cfdataset and cfdataset.cb.cf_role is None:
            logging.debug('adding a dummy profile variable as none was '
                          'provided')
            cfdataset['profile'] = xr.DataArray(
                dims=(),
                attrs={'cf_role': 'profile_id'},
                data=np.empty((), dtype=np.int)
            )

        for _, v in cfdataset.data_vars.items():
            v.encoding['coordinates'] = 'time lat lon z'

        return cfdataset

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[DiscreteFeature]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if cls._unexpected_featureType(cfdst) or \
                cls._unexpected_cdm_data_type(cfdst):
            return

        if len(dataset.cb.cf_instance_dims) != 0:
            return

        # if dataset.cb.instance_dim is not None:
        #     profile = dataset.cf.instance_dim
        #     if not (
        #         cls._check_coord_dims(dataset, 'latitude', profile, 1) and
        #         cls._check_coord_dims(dataset, 'latitude', profile, 1) and
        #         cls._check_coord_dims(dataset, 'time', profile, 1) and
        #         cls._check_coord_dims(dataset, dataset.cb.vertical,
        #                               dataset.cb.cf_axis_dims['Z'], None)
        #     ):
        #         return

        for coord in ['longitude', 'latitude', 'time']:
            if not cls._check_coord_dims(dataset, coord, []):
                return
        if not cls._check_coord_dims(
                dataset, 'vertical', dataset.cb.cf_axis_dims['Z'],None):
            return

        return Profile



#!/usr/bin/env python
# coding=utf-8
"""
"""
import datetime
from .exceptions import *
from .helpers import *

import numpy as np

# import features
from cerbere.feature.cbasefeature import BaseFeature
from cerbere.feature.cgrid import Grid
from cerbere.feature.cimdcollection import IMDCollection
from cerbere.feature.comdcollection import OMDCollection
from cerbere.feature.cpoint import Point
from cerbere.feature.cprofile import Profile
from cerbere.feature.cracollection import CRACollection
from cerbere.feature.cswath import Swath
from cerbere.feature.ctimeseries import TimeSeries
from cerbere.feature.ctimeseriesprofile import TimeSeriesProfile
from cerbere.feature.ctrajectory import Trajectory
from cerbere.feature.ctrajectoryprofile import TrajectoryProfile


# preserve the variable attributes during xarray operations
xr.set_options(keep_attrs=True)


def dt64todatetime(dt64):
    unix_epoch = np.datetime64(0, 's')
    one_second = np.timedelta64(1, 's')
    seconds_since_epoch = (dt64 - unix_epoch) / one_second

    if isinstance(seconds_since_epoch, xr.DataArray):
        seconds_since_epoch = seconds_since_epoch.data.item()

    return datetime.datetime.utcfromtimestamp(seconds_since_epoch)



"""
Equivalent terms commonly found in Earth Observation data files from various
providers for Cerbere / CF terms
"""

# common matching dimensions and fields in netCDF files

DIM_MATCHING = {
    'time': 'T',
    'lon': 'X',
    'longitude': 'X',
    'lat': 'Y',
    'latitude': 'Y',
    'x': 'X',
    'y': 'Y',
    'mes': 'station',
    'station': 'station',
    'ni': 'X',
    'cell': 'X',
    'ra_size': 'X',
    'col': 'X',
    'nj': 'Y',
    'row': 'Y',
    'az_size': 'Y',
    'rows': 'Y',
    'columns': 'X',
    'NUMROWS': 'Y',
    'NUMCELLS': 'X',
    'across_track': 'X',
    'along_track': 'Y',
    'TIME': 'T',
    'LATITUDE': 'Y',
    'LONGITUDE': 'X',
    'z': 'Z',
}

COORD_ALIASES = {
    'longitude': {'lon', 'longitude', 'LONGITUDE'},
    'latitude': {'lat', 'latitude', 'LATITUDE'},
    'time': {'TIME'},
    'vertical': {'z'},
}

TIME_COVERAGE_ATTRS = {
    'date': {
        'start': 'start_date',
        'end': 'stop_date'
    },
    'time': {
        'start': 'start_time',
        'end': 'stop_time'
    },
    'time_coverage': {
        'start': 'time_coverage_start',
        'end': 'time_coverage_end'
    },
    'time_coverage2': {
        'start': 'time_coverage_start',
        'end': 'time_coverage_stop'
    },
    'meas_time': {
        'start': 'first_meas_time',
        'end': 'last_meas_time'
    },
}

BBOX_CORNERS = {
    'latmin': [
        'southernmost_latitude', 'geospatial_lat_min', 'south_latitude'
    ],
    'latmax': [
        'northernmost_latitude', 'geospatial_lat_max', 'north_latitude'
    ],
    'lonmin': [
        'westernmost_longitude', 'geospatial_lon_min', 'west_longitude'
    ],
    'lonmax': [
        'easternmost_longitude', 'geospatial_lon_max', 'east_longitude'
    ]
}


# -*- coding: utf-8 -*-
"""
Feature class for orthogonal multidimensional collection of features.
"""
from __future__ import absolute_import, division, print_function

from typing import Tuple

import xarray as xr

from cerbere.feature.feature import Feature

__all__ = ['OMDCollection']


class OMDCollection(Feature):
    """
    Feature class for an orthogonal multidimensional collection of features.

    A :class:`~cerbere.feature.omdcollection.OMDCollection` class is defined by
    geolocation coordinates field and geolocation dimension of the feature, plus
    the instance dimension defined for this feature.

    For instance, for a :class:`~cerbere.feature.profile.Profile` feature, the
    instance dimension is ``profile``; the CF convention states that the
    geolocation coordinates field should be defined as: ``lon(profile)``,
    ``lat(profile)``, ``time(profile)``, ``z(z)``; observation field should be
    defined as ``fieldname(profile, z)``.
    """
    def __init__(
            self,
            *args,
            feature_class: 'Feature' = None,
            **kwargs):
        """
        Args:
            feature_class (Feature) : the feature class which the collection is
                made of.

        See:
        Feature
        """
        if feature_class is None:
            raise ValueError(
                'A value must be provided for ``feature_class`` arg'
            )
        self.feature_class = feature_class
        super(OMDCollection, self).__init__(*args, **kwargs)

        self._std_dataset.attrs['featureType'] = feature_class._instance_dimname

    @property
    def _feature_geodimnames(self):
        return (self.feature_class._instance_dimname,) + \
               self.feature_class._feature_geodimnames

    def get_geocoord_dimnames(
            self, fieldname: str,
            values: 'xr.DataArray') -> Tuple[str, ...]:
        if fieldname in ['depth', 'height']:
            return self.feature_class._feature_geodimnames
        else:
            return self.feature_class._instance_dimname



from pathlib import Path
import re
import typing as T

import numpy as np


class GHRSST:

    pattern: str = re.compile(r"^[0-9]{8,14}-.*-L[2P|3\w|4]P*_GHRSST-.*nc$")
    engine: str = "netcdf4"
    description: str = "Use GHRSST files in Xarray"
    url: str = "https://link_to/your_backend/documentation"

    @staticmethod
    def guess_can_open(filename_or_obj: T.Union[str, Path]) -> bool:
        return re.match(
            GHRSST.pattern,
            Path(filename_or_obj).name) is not None

    @staticmethod
    def postprocess(ds, decode_times: bool = True):
        if "sst_dtime" in ds.variables and decode_times:
            # replace with new time variable decoding GHRSST time
            dunit = ds.sst_dtime.attrs.get(
                "units", ds.sst_dtime.encoding.get("units"))
            if dunit is None:
                raise ValueError("missing unit for sst_dtime")
            if dunit not in ["second", "seconds"]:
                raise TypeError("Can not process sst_dtime in {}".format(dunit))
            reference = np.datetime64(ds["time"].values[0])

            time = ds["sst_dtime"].astype("timedelta64[s]") + reference
            time.attrs = ds["time"].attrs
            time.encoding = ds["time"].encoding

            # remove dummy time dimension
            ds = ds.squeeze(dim="time")

            # replace with single time field
            ds = ds.drop_vars(["time", "sst_dtime"])
            ds['time'] = time.squeeze(dim="time")
            ds = ds.set_coords(["time"])
            if "comment" in ds["time"].attrs:
                ds["time"].attrs.pop("comment")
            ds["time"].attrs["long_name"] = "measurement time"

        # normalize collection id (GDS 1.x => 2.x)
        if "DSD_entry_id" in ds.attrs:
            ds.attrs["id"] = ds.attrs.pop("DSD_entry_id")

        return ds



# -*- coding: utf-8 -*-

"""
cerbere.exceptions
~~~~~~~~~~~~~~~~~~~
This module contains the set of Cerbere exceptions.
"""


class AxisError(Exception):
    """Exception class for CF axis coordinates management"""
    pass


class BadAxisError(AxisError):
    pass


class MissingAxisError(AxisError):
    pass


class CoordinateError(Exception):
    """Exception class for CF spatial and temporal coordinates management"""
    pass


class FeatureTypeError(Exception):
    pass


# warnings
class CWarning(UserWarning):
    """Cerbere warning class"""
    pass



# -*- coding: utf-8 -*-
"""
Feature class for a set of randomly sampled points.
"""
import typing as T

import xarray as xr

from cerbere.feature.cdiscretefeature import DiscreteFeature

__all__ = ['Point']


class Point(DiscreteFeature):
    """
    To represent data at scattered locations and times with no implied
    relationship among of coordinate positions, both data and coordinates
    must share the same (sample) instance dimension. Because each feature
    contains only a single data element, there is no need for a separate
    element dimension. The representation of point features is a special,
    degenerate case of the standard four representations. The coordinates
    attribute is used on the data variables to unambiguously identify the
    relevant space and time auxiliary coordinate variables.

    A :class:`~cerbere.feature.point.Point` feature is defined in CF convention
    by geolocation coordinates field: ``lon(obs)``, ``lat(obs)``,
    ``time(obs)``, optionally ``z(obs)``, all having a single geolocation
    dimension: ``obs``.
    """
    # must be defined for any feature corresponding to a CF discrete sampling
    # geometry, Nobe otherwise
    cf_feature_type = 'point'

    # the instance dimension name for collection of features
    cf_instance_axis = 'obs'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'longitude': (cf_instance_axis,),
        'latitude': (cf_instance_axis,),
        'time': (cf_instance_axis,),
        'vertical': (cf_instance_axis,),
        'data': (cf_instance_axis,)
    }

    # optional axes for the feature
    cf_optional_axes = ['vertical']

    # Unidata Common Data Model provides a wider set of feature types than CF
    cdm_data_type = 'point'

    def __init__(self, *args, **kwargs):
        super(Point, self).__init__(*args, **kwargs)

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[DiscreteFeature]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if cls._unexpected_featureType(cfdst) or \
                cls._unexpected_cdm_data_type(cfdst):
            return

        # check lat/lon/time/z are zero dimensional
        for coord in ['longitude', 'latitude', 'time', 'vertical']:
            if not cls._check_coord_dims(dataset, coord,[]):
                return

        # check no axis coord
        if dataset.cb.cf_axis_coords != {}:
            return

        return Point
        #
        # if not (set(cfdst['lon'].dims) == set(cfdst['lat'].dims) == set(
        #         cfdst['time'].dims)):
        #     return
        # if cfdst.cb.Z is not None and set(cfdst['lon'].dims) != set(
        #         cfdst.cb.Z.dims):
        #     return
        # if len(set(cfdst['lon'].dims)) != 1 or cfdst['lon'].dims[0] == 'time':
        #     return
        #
        # # check provided axes if any
        # if cfdst.cb.X is not None and cfdst.cb.X.name != 'lon':
        #     return
        # if cfdst.cb.Y is not None and cfdst.cb.Y.name != 'lat':
        #     return
        # if cfdst.cb.T is not None and cfdst.cb.T.name != 'time':
        #     return
        #
        # return Point



"""
Base class for unitary testing of cerbere feature classes with pytest.
"""
from datetime import datetime
import logging
from pathlib import Path

import pytest

import numpy as np
import xarray as xr

import cerbere
from cerbere.feature.cbasecollection import BaseCollection


logger = logging.getLogger()
logger.setLevel('DEBUG')


@pytest.fixture(scope='module')
def output_dir(tmpdir_factory):
    return tmpdir_factory.mktemp('output')


@pytest.fixture(scope='module')
def instance_class():
    return


@pytest.fixture
def input_file(output_dir, feature_instance):

    # creates a test xarray object
    feat = feature_instance

    # save it to disk
    fname = Path(output_dir) /'test_feature_dataset.nc'
    if fname.exists():
        fname.unlink()
    feat.dataset.cb.save(fname)

    return Path(output_dir) /'test_feature_dataset.nc'


@pytest.fixture
def output_file(output_dir):
    return Path(output_dir) / 'test_saved_feature.nc'


@pytest.fixture
def test_var():
    return 'my_var'


@pytest.fixture
def test_create_feature_from_dict(feature_class, feature_instance):
    basefeat = feature_instance
    feat = feature_class(xr.Dataset.from_dict({
        'coords': {
            'lat': {
                'data': basefeat.dataset.lat.values,
                'dims': basefeat.dataset.lat.dims
            },
            'lon': {
                'data': basefeat.dataset.lon.values,
                'dims': basefeat.dataset.lon.dims
            },
            'time': {
                'data': basefeat.dataset.time.values,
                'dims': basefeat.dataset.time.dims
            }
        },
        'data_vars':{
            'my_var': {
                'data': basefeat.dataset['my_var'].values,
                'dims': basefeat.dataset['my_var'].dims
            }
        },
        'attrs': basefeat.dataset.attrs}))

    assert isinstance(feat, feature_class)


def test_create_feature_from_xarray(
        feature_class, feature_instance, instance_class):

    i_class_name = None
    if instance_class is not None:
        i_class_name = instance_class.__name__

    feat = feature_class(
        feature_instance.dataset, instance_class=i_class_name)

    assert isinstance(feat, feature_class)


def test_init_from_ncfile(input_file, feature_class, instance_class):
    dst = xr.open_dataset(input_file)

    i_class_name = None
    if instance_class is not None:
        i_class_name = instance_class.__name__

    feat = feature_class(dst, instance_class=i_class_name)

    assert isinstance(feat, feature_class)
    assert feat.dataset.cb.time_coverage_start is not None


def test_save_feature(feature_instance, output_file):
    feat = feature_instance
    feat.dataset.cb.save(path=output_file)


def test_get_lat(feature_instance):
    data = feature_instance.dataset['lat'].cb.to_masked_array()

    assert isinstance(data, np.ndarray)


def test_get_lon(feature_instance):
    data = feature_instance.dataset['lon'].cb.to_masked_array()

    assert isinstance(data, np.ndarray)


def test_get_times(feature_instance):
    data = feature_instance.dataset['time'].cb.to_masked_array()

    assert isinstance(data, np.ndarray)
    assert np.issubdtype(data.dtype, np.datetime64)


def test_get_values(feature_instance, test_var):
    feat = feature_instance
    data = feat.dataset[test_var].cb.to_masked_array()

    assert isinstance(data, np.ma.MaskedArray)


def test_get_geocoord(feature_instance):
    assert 'lat' in feature_instance.dataset.coords


def test_cf_dims(feature_instance, instance_class):
    """verify the expected dimensions for this feature are present"""
    dims = feature_instance.ds.cb.cf_dims

    # if feature_instance.featureType is None:
    #     return

    # if isinstance(feature_instance, BaseCollection):
    #     assert feature_instance.instance_class == instance_class
    #     cf_canonical_dims = feature_instance.instance_class.cf_canonical_dims
    # else:
    #     cf_canonical_dims = feature_instance.cf_canonical_dims
    cf_canonical_dims = feature_instance.cf_canonical_dims

    print(feature_instance.ds)
    print(feature_instance.ds.encoding)

    for var_type, c_dims in cf_canonical_dims.items():
        if var_type == 'data':
            continue

        coord = feature_instance.ds.cb.cf_axis_coords.get(var_type.upper())
        if coord is None:
            assert coord not in feature_instance.ds.cb.cf_coords
            continue

        # replace generic axes in v with actual name
        coord_dims = feature_instance.ds.cb.cfdataset[coord].dims
        c_dims = [feature_instance.ds.cb.cf_axis_dims.get(_, _) for _ in c_dims]

        assert (
            set(coord_dims) == set(c_dims) or
            set(coord_dims) == set(c_dims) - {feature_instance.cf_instance_axis}
        )


#
# def test_geodimnames(feature_instance, feature_dimnames):
#     dims = feature_instance.dataset.cb.geodimnames
#
#     assert isinstance(dims, tuple)
#     print(dims, feature_dimnames)
#     print(feature_instance)
#     #assert all([dims[i] == feature_dimnames[i] for i in range(len(dims))])
#     assert set(dims) == set(feature_dimnames)
#
#
# def test_geodimsizes(feature_instance, feature_dimsizes):
#     sizes = feature_instance.dataset.cb.geodimsizes
#
#     assert isinstance(sizes, tuple)
#     print(sizes)
#     print(feature_instance.dataset)
#     assert all([sizes[i] == feature_dimsizes[i] for i in range(len(sizes))])


def test_clone_field(input_file, feature_class, feature_instance, test_var):
    feat = feature_class(xr.open_dataset(input_file))
    cloned = feat.dataset[test_var].copy(deep=True)

    print(cloned)
    assert isinstance(cloned, xr.DataArray)

    feat2 = feat = feature_instance
    cloned2 = feat2.dataset[test_var].copy(deep=True)
    print(cloned2)
    assert isinstance(cloned2, xr.DataArray)
    assert cloned.name == cloned2.name


def test_guess_feature(input_file, feature_class):
    dst = xr.open_dataset(input_file)
    print(cerbere.guess_feature(dst), feature_class)
    assert cerbere.guess_feature(dst) == feature_class


def test_feature_iteration(feature_instance):
    if isinstance(feature_instance, BaseCollection):
        feature = feature_instance[0]
        assert isinstance(feature, feature_instance.instance_class)



# -*- coding: utf-8 -*-
"""
Feature class for the profile observation pattern.
"""
from __future__ import absolute_import, division, print_function

from typing import Tuple

import xarray as xr

from cerbere.feature.feature import Feature

__all__ = ['Profile']


class Profile(Feature):
    """
    Feature class for the profile observation patterns.

    A :class:`~cerbere.feature.profile.Profile` feature is defined by
    geolocation coordinates field: ``lon(profile)``, ``lat(profile)``,
    ``time(profile)``, ``z(profile, z)``, all having a single geolocation
    dimension: ``profile``.
    """
    _instance_dimname = 'profile'
    _feature_geodimnames = 'z',

    def __init__(self, *args, **kwargs):
        super(Feature, self).__init__(*args, **kwargs)

        # CF specific attrs and vars for profile feature
        self._std_dataset['profile'] = xr.DataArray(
            dims=(),
            attrs={'cf_role': 'profile_id'}
        )
        self._std_dataset.attrs['featureType'] = 'profile'
        for _, v in self._std_dataset.data_vars.items():
            v.attrs['coordinates'] = 'time lat lon z'

    def get_geocoord_dimnames(
            self, fieldname: str,
            values: 'xr.DataArray') -> Tuple[str, ...]:
        if fieldname in ['depth', 'height']:
            return 'z',
        else:
            return ()



# -*- coding: utf-8 -*-
"""
Feature class for the time series observation pattern at a fix station.
"""
from __future__ import absolute_import, division, print_function

from typing import Tuple

from cerbere.dataset.dataset import STANDARD_Z_FIELD
from cerbere.feature.feature import Feature

__all__ = ['TimeSeries']


class TimeSeries(Feature):
    """
    Feature class for the time series observation patterns.

    A :class:`~cerbere.feature.timeseries.TimeSeries` feature is defined by
    geolocation coordinates field: ``lon()``, ``lat()``,
    ``time(time)``, ``depth/alt ()``, none having a geolocation dimension
    (except ``time`` for T axis coordinate).
    """
    _instance_dimname = 'station'
    _feature_geodimnames = 'time',

    def __init__(self, *args, **kwargs):
        super(TimeSeries, self).__init__(*args, **kwargs)

        # CF specific attrs and vars for profile feature
        self._std_dataset.attrs['featureType'] = 'timeSeries'

        # mandatory coordinates
        for _, v in self._std_dataset.data_vars.items():
            v.attrs['coordinates'] = 'time lat lon z'

    def get_geocoord_dimnames(
            self, fieldname: str, *args, **kwargs) -> Tuple[str, ...]:
        if (fieldname in STANDARD_Z_FIELD + ['lat', 'lon'] or
                self.coords[fieldname].is_axis('Z')):
            return ()
        else:
            return 'time',



"""
Helper functions to save data to Cerbere CF / ACDD convention
"""
from datetime import datetime
import logging
from pathlib import Path
import typing as T

import pandas as pd
import shapely.geometry
import numpy as np
import xarray as xr

import cerbere.cfconvention as cf


def to_netcdf(
        path: T.Union[str, Path],
        dataset: xr.Dataset,
        profile: str = 'default_saving_profile.yaml',
        force_profile: bool = False,
        keep_src_encoding: bool = False,
        **kwargs):
    """
    Save a dataset as CF compliant netCDF file

    Args:
        path:
        dataset:
        profile (str, optional): path to a specific format profile to
            apply before saving (or default formatting profile is used).
        force_profile (bool, optional): force profile attribute values to
            supersede existing ones in dataset attributes.
        keep_src_encoding (bool): keep original dtype, _FillValue
            and scaling if any (through `add_offset` or `scale_factor`
            attributes) as in the source data.

    Returns:

    """
    if 'format' in kwargs and 'NETCDF' not in kwargs['format']:
        raise ValueError(f'Unknown output format : {kwargs["format"]}')

    # saved version of the dataset
    dataset = dataset.copy(deep=False)

    # clear previous formatting rules
    if not keep_src_encoding:
        for v in dataset.variables:
            for attr in ['dtype', 'scale_factor', 'add_offset']:
                if attr in dataset[v].encoding:
                    dataset[v].encoding.pop(attr)

    # apply new formatting rules
    saved_dataset = _convert_format(dataset,
        profile=profile, force_profile=force_profile)

    # ensure proper type in output attributes so that it does not
    # generate errors for types not managed by the NetCDF lib
    _format_attrs_for_nc(saved_dataset)

    def has_scaling(variable: str):
        acc = dataset[variable].cb
        return acc.scale_factor is not None or acc.add_offset is not None

    # ensure original or overriding encoding
    for v in saved_dataset.variables:
        svar = saved_dataset[v]
        if keep_src_encoding:
            encoding = dataset.cb._orig_dataset[v].encoding
        else:
            encoding = svar.encoding

        if 'zlib' not in encoding:
            encoding['zlib'] = True
        if 'complevel' not in encoding:
            encoding['complevel'] = 4

        # save a _FillValue matching the encoding data type
        if 'grid_mapping_name' not in svar.attrs:
            encoding['_FillValue'] = encoding.get(
                '_FillValue', svar.attrs.get('_FillValue'))
        else:
            if '_FillValue' in encoding:
                encoding.pop('_FillValue')

        if np.issubdtype(svar.dtype, np.datetime64) \
                and svar.cb.units is None:
            svar.cb.units = cf.DEFAULT_TIME_UNITS

        if (np.issubdtype(svar.dtype, np.datetime64)
                or np.issubdtype(svar.dtype, np.timedelta64)):
            dtype = np.dtype(np.float64)
        else:
            dtype = encoding.get('dtype', svar.dtype)
        encoding['dtype'] = dtype

        # adjust missing value attribute types if packing is applied
        for matt in ['valid_min', 'valid_max', 'valid_range']:
            if matt in svar.attrs:
                svar.attrs[matt] = np.array(
                    svar.attrs[matt],
                    dtype=dtype)

        saved_dataset[v].encoding = encoding

        if np.issubdtype(saved_dataset[v].dtype, np.datetime64):
            # _FillValue can not be NaT or netcdf writing will fail
            if pd.isnull(saved_dataset[v].encoding['_FillValue']):
                saved_dataset[v].encoding['_FillValue'] = 1e20

            # units must be in encoding
            if 'units' in saved_dataset[v].attrs:
                saved_dataset[v].encoding['units'] = saved_dataset[
                    v].attrs.pop('units')

    saved_dataset.to_netcdf(
        path=path,
        engine='h5netcdf',
        **kwargs
    )


def _add_global_attrs(
        dataset: xr.Dataset,
        attrs: T.Mapping[str, T.Any],
        force_profile: bool = False) -> None:
    """Add default attributes to the dataset from a attribute definition
    file.

    Used when formatting with a template.

    Args:
        attrs (dict): the global attributes to add when saving the dataset
    """
    # add attributes
    # don't override previous attribute values
    for att in attrs:
        if att not in dataset.attrs or force_profile:
            if attrs[att] is None:
                continue
            dataset.attrs[att] = attrs[att]


def _add_var_attrs(
        dataset: xr.Dataset,
        attrs: T.Mapping[str, T.Any],
        force_profile: bool = False) -> None:
    """Add field attributes to the dataset from a attribute definition
    file.

    Used when formatting with a template.

    Args:
        attrs (dict): the field attributes to add when saving the dataset
    """
    # add attributes
    # don't override previous attribute values
    for v in attrs:
        if v not in dataset.variables:
            logging.warning(
                f'Variable {v} not found in the dataset to save. Skipping')
            continue

        for att in attrs[v]:
            if att not in dataset.variables[v].attrs or force_profile:
                value = attrs[v][att]
                if value is None:
                    continue
                dataset.variables[v].attrs[att] = value


def _add_var_encoding(
        dataset: xr.Dataset,
        attrs: T.Mapping[str, T.Any],
        force_profile: bool = False) -> None:
    """Add field encoding to the dataset from a attribute definition
    file.

    Used when formatting with a template.

    Args:
        attrs (dict): the field encoding attributes to add when saving the
            dataset
    """
    # add attributes
    # overrides previous attribute values!!
    for v in attrs:
        if v not in dataset.variables:
            logging.warning(
                f'Variable {v} not found in the dataset to save. Skipping')
            continue

        if attrs[v] is None:
            continue

        for att in attrs[v]:
            if att not in dataset.variables[v].encoding or force_profile:
                value = attrs[v][att]
                if value is None:
                    continue
                dataset.variables[v].encoding[att] = value


def _convert_format(cfdataset, profile=None, force_profile=False, **kwargs):
    """Implement specific formatting rules to a dataset.

    Used before saving the dataset to match some specific format
    requirements when writing the dataset on file.

    A custom format profile can be provided, using ``profile``
    argument, otherwise the default cerbere profile will be used.

    Refer to :func:`~cerbere.cfconvention.default_profile`
    """
    dataset = cfdataset.copy()

    # read custom settings for the dataset saving
    gattrs, vattrs, encoding = cf.default_profile(profile=profile)

    # add global attributes from custom settings
    _add_global_attrs(dataset, gattrs, force_profile=force_profile)

    # add field attributes from custom settings
    _add_var_attrs(dataset, vattrs, force_profile=force_profile)

    # add field encoding from custom settings
    _add_var_encoding(dataset, encoding, force_profile=force_profile)

    return dataset


def _format_attrs_for_nc(dataset):
    """format the attributes in an acceptable type for netCDF"""
    # global attrs
    invalid_attrs = []
    for att in dataset.attrs:
        # remove None attributes
        if dataset.attrs[att] is None:
            invalid_attrs.append(att)
            continue
        # convert datetime objects to string
        dataset.attrs[att] = _format_attr(dataset.attrs[att])
    for att in invalid_attrs:
        dataset.attrs.pop(att)

    # variable attrs
    for varname in dataset._variables:
        var = dataset._variables[varname]
        invalid_attrs = []
        for att in var.attrs:
            if var.attrs[att] is None:
                invalid_attrs.append(att)
                continue

            if att == '_FillValue':
                # should be in encoding
                if '_FillValue' in var.encoding:
                    raise ValueError(f'two fill values defined for {varname}')
                var.encoding[att] = var.attrs.pop(att)

        for att in invalid_attrs:
            var.attrs.pop(att)


def _format_attr(attrval):
    if isinstance(attrval, datetime):
        return attrval.strftime(cf.FORMAT_TIME)
    elif attrval is None:
        return ''
    elif isinstance(attrval, shapely.geometry.base.BaseGeometry):
        return attrval.wkt
    return attrval


# -*- coding: utf-8 -*-
"""
BackendEntrypoint class for KNMI / OSI SAF scatterometer Level 2 NETCDF files
"""
import os

import numpy as np
from xarray import open_dataset
from xarray.backends import BackendEntrypoint


class OSISAFSCATL2(BackendEntrypoint):
    def open_dataset(
        self,
        filename_or_obj,
        mask_and_scale=True,
        decode_times=True,
        decode_timedelta=True,
        use_cftime=True,
        concat_characters=True,
        decode_coords=True,
        drop_variables=None,
        # other backend specific keyword arguments
        # `chunks` and `cache` DO NOT go here, they are handled by xarray
    ):
        ds = open_dataset(
            filename_or_obj, mask_and_scale=mask_and_scale,
            decode_times=True, decode_timedelta=decode_timedelta,
            use_cftime=False, concat_characters=concat_characters,
            decode_coords=decode_coords, drop_variables=drop_variables)

        ds.cb.cerberize(
            dim_matching={
                'numrows': 'row',
                'NUMROWS': 'row',
                'numcells': 'cell',
                'NUMCELLS': 'cell',
                'NUMAMBIGS': 'solutions',
                'NUMVIEWS': 'views'
            },
            var_matching={
                'row_time': 'time',
                'wvc_lat': 'lat',
                'wvc_lon': 'lon'
            }
        )

        return ds.cb.cfdataset

    open_dataset_parameters = [
        "filename_or_obj", "mask_and_scale", "decode_times",
        "decode_timedelta", "use_cftime", "concat_characters", "decode_coords"]

    def guess_can_open(self, filename_or_obj):
        try:
            _, ext = os.path.splitext(filename_or_obj)
        except TypeError:
            return False
        return ext in {".nc"}

    description = "Use OSI SAF NetCDF scatterometer files in Xarray and cerbere"

    url = "https://link_to/your_backend/documentation"



# -*- coding: utf-8 -*-
"""
Classes for the handling the fields.
"""
from __future__ import absolute_import, division, print_function

import collections
import copy
import logging
from typing import Any, Hashable, List, Mapping, Optional, Tuple, Type, Union

try:
    # python >= 3.7.2
    from typing import OrderedDict
except ImportError:
    # python >= 3.6
    from typing import MutableMapping as OrderedDict

import warnings

import numpy as np
import xarray as xr

import cerbere.cfconvention as cf
import cerbere

__all__ = ['Field']


FIELD_ATTRS = [
    'standard_name',
    'authority',
    'units',
    'valid_min',
    'valid_max',
    '_FillValue'
]
FIELD_EXCL_ATTRS = [
    'long_name'
]


class Field(object):
    """A Field describes a scientific data array. It contains data and
    metadata attributes.

    This is an extension of xarray's DataArray with stricter requirements on
    attributes and the encoding of masked arrays.

    A :class:`Field` object can be attached to a  :class:`~cerbere.dataset.dataset.Dataset`
    object (of any class inherited from a :class:`~cerbere.dataset.dataset.Dataset`,
    provided with the ``dataset`` argument.

    Args:
        data: the scientific data. Can be of the following types:

              * :class:`np.ma.MaskedArray`
              * :class:`xarray.DataArray`

            Optional (the field can be created with default values)

        name: the label of the field (don't use any white space). This
            corresponds to the variable name in a netcdf file

        dims (tuple or OrderedDict): the scientific array dimension names. If no
            data are provided (new masked field), provide dimension names and
            sizes as a dictionary.

        dtype: the type of the data. Infer the type from the provided data by
            default. xarray changes the data type of arrays when using masking,
            defining the field true type with this keyword will ensure you
            always get the data in the expected dtype when accessing with
            `get_values` method. Data will also be saved with the expected
            dtype.

        precision (int, optional): number of significant digits (when writing
            the data on file)

        fields : the subfields composing the main field.
            This is intended to group for instance components of the same
            variable (such as a vector's northward and eastward components
            for wind, currents,...). This allows to relate these components
            to the same physical variable (e.g wind) and to a single
            qc_levels and qc_details information.

        fill_value : the default value to associate with missing data in the
            field's values. The fill value must be of the same type as
            `dtype`. If None, default fill value will be used
            unless `no_missing_value` is set.

        no_missing_value: if True, the field will not contain any missing value.
            In practice no _FillValue attribute will be written for this field
            when saving it. Use this for mask or flag fields for instance.

        attrs (dict) : a dictionary of the metadata associated with the field's
            values.
        units : the units in which the data are given (if
            applicable)

        description (str) : full name of the phenomenon. This corresponds
            to a long_name in attribute in a netCDF file.

        authority (str): naming authority referencing the provided
            standard name

        standard_name (optional, str): standard label for a phenomenon, with
            respect to the convention stated in `authority` argument.
            This corresponds to a standard_name attribute in a CF compliant
            NetCDF file.

        quality_vars: list of related quality fields
    """
    def __init__(self,
                 data,
                 name: Optional[Hashable] = None,
                 dims: Optional[
                     Union[Tuple[str, ...], OrderedDict[Hashable, int]]] = None,
                 dtype: Optional[Union[np.dtype, Type[float]]] = None,
                 components: Optional[Tuple['Field']] = None,
                 dataset: Optional['Dataset'] = None,
                 fill_value: Optional[Any] = None,
                 no_missing_value: bool = False,
                 precision: Optional[int] = None,
                 description: Optional[str] = None,
                 standard_name: Optional[Union[str, Tuple[str, str]]] = None,
                 units: Optional[str] = None,
                 quality_vars: Optional[List[str]] = None,
                 attrs: Optional[Mapping[str, Any]] = None,
                 encoding: Optional[Mapping[str, Any]] = None,
                 **kwargs) -> None:
        """
        """
        self._array = cerbere.new_array(
            data,
            name=name,
            dims=dims,
            science_dtype=dtype,
            fill_value=fill_value,
            precision=precision,
            long_name=description,
            standard_name=standard_name,
            units=units,
            ancillary_variables=quality_vars,
            attrs=attrs,
            **kwargs
        )
        if encoding is not None:
            self._array.cerbere.cfdataset.encoding.update(encoding)

        if no_missing_value:
            self._array.cerbere.fill_value = None

    @classmethod
    def to_field(cls, data: xr.DataArray) -> 'Field':
        """Cast a xarray DataArray to a
        :class:`cerbere.datamodel.field.Field` object
        """
        return Field(data=data)

    def to_xarray(
            self,
            silent: bool = False,
            decoding: bool = False) -> xr.DataArray:
        """
        Return the field values a xarray DataArray

        Args:
            decoding: return the DataArray in its scientific dtype instead of
                its DataArray dtype.
        """
        if decoding:
            return self._array.cerbere.cfarray.astype(
                self._array.cerbere.science_dtype)

        return self._array.cerbere.cfarray

    def __str__(self):
        return self._array.cerbere.__str__()

    @property
    def name(self) -> Hashable:
        """Name of the field"""
        return self._array.cerbere.cfarray.name

    @name.setter
    def name(self, value: Hashable):
        self._array.cerbere.cfarray.name = value

    @property
    def attrs(self) -> Mapping[Hashable, Any]:
        """A dict of the field attributes"""
        return self._array.cerbere.cfarray.attrs

    @attrs.setter
    def attrs(self, attrs: Mapping[Hashable, Any]):
        self._array.cerbere.cfarray.attrs = attrs

    @property
    def encoding(self) -> Mapping[Hashable, Any]:
        """A dict of the field encoding attributes"""
        return self._array.cerbere.cfarray.encoding

    @encoding.setter
    def encoding(self, attrs: Mapping[Hashable, Any]):
        self._array.cerbere.cfarray.encoding = attrs

    @property
    def sizes(self) -> Mapping[Hashable, int]:
        """A tuple of the field dimensions name and size"""
        if self.dataset is None:
            return self._array.cerbere.cfarray.sizes
        else:
            return self.dataset.get_field_sizes(self.name)

    @property
    def dims(self) -> Tuple[Hashable]:
        """A tuple of the field dimensions name and size"""
        return self._array.cerbere.cfarray.dims

    @dims.setter
    def dims(self, dims):
        self._array.cerbere.cfarray.dims = dims

    @property
    def dimnames(self) -> Tuple[Hashable]:
        """Tuple of the field's dimension names"""
        return self.dims

    def get_dimsize(self, dimname) -> int:
        """Return the size of a field dimension"""
        return self._array.cerbere.cfarray.sizes[dimname]

    @property
    def fill_value(self) -> Union[Any, None]:
        """The value for missing data in the field"""
        return self._array.cerbere.fill_value

    @fill_value.setter
    def fill_value(self, fill_value: Any):
        """set the value for missing data"""
        self._array.cerbere.fill_value = fill_value

    @property
    def precision(self) -> Union[int, None]:
        """The least significant digit for the values of the field.

        Used for storage optimization when writing on disk. The least
        significant digit is the power of ten of the smallest decimal place in
        the data that is a reliable value.
        """
        return self._array.cerbere.precision

    @precision.setter
    def precision(self, precision: int):
        """Set the least significant digit for the values of the field.

        Used for storage optimization when writing on disk. The least
        significant digit is the power of ten of the smallest decimal place in
        the data that is a reliable value.
        """
        self._array.cerbere.precision = precision

    @property
    def valid_min(self) -> Union[Any, None]:
        """The minimum valid value in the field data"""
        return self._array.cerbere.valid_min

    @valid_min.setter
    def valid_min(self, value: Any):
        """set the minimum valid value"""
        self._array.cerbere.valid_min = value

    @property
    def valid_max(self) -> Union[Any, None]:
        """The maximum valid value in the field data"""
        return self._array.cerbere.valid_max

    @valid_max.setter
    def valid_max(self, value: Any):
        """set the maximum valid value"""
        self._array.cerbere.valid_max = value

    @property
    def units(self) -> Union[str, None]:
        """The field data units (``units`` CF attribute)"""
        return self._array.cerbere.units

    @units.setter
    def units(self, units: str):
        """Set the field data units (``units`` CF attribute)"""
        self._array.cerbere.units = units

    @property
    def description(self) -> Union[str, None]:
        """The field description (``long_name`` CF attribute)"""
        return self._array.cerbere.long_name

    @description.setter
    def description(self, description: str) -> None:
        """set the field description (``long_name`` CF attribute)"""
        self._array.cerbere.long_name = description

    @property
    def standard_name(self) -> Union[str, None]:
        """The field standard name (``standard_name`` CF attribute)"""
        return self._array.cerbere.standard_name

    @standard_name.setter
    def standard_name(self, standard_name: str) -> None:
        """set the standard_name (``standard_name`` CF attribute)"""
        if isinstance(standard_name, tuple):
            self._array.cerbere.standard_name = standard_name[0]
            self._array.cerbere.authority = standard_name[1]
        elif standard_name is not None:
            self._array.cerbere.standard_name = standard_name
            self._array.cerbere.authority = cf.CF_AUTHORITY

    @property
    def components(self):
        """subfield of a multi-array field (such as vector, broken down into
        u and v fields)"""
        raise DeprecationWarning

    @name.setter
    def components(self, value):
        raise DeprecationWarning

    @property
    def dtype(self) -> np.dtype:
        """The type of the field data"""
        return self._array.cerbere.science_dtype

    def is_composite(self) -> bool:
        """
        True if the field is a composite field.

        A composite field is a composition of sub fields (vector components,
        real and imaginary part of a complex, ...)
        """
        raise DeprecationWarning

    def is_axis(self, axis: str) -> bool:
        """Returns true if the field is the coordinate field for the given axis.

        Args:
            axis: axis name (among X, Y, Z, T)
        """

    def get_values(
            self,
            index: Optional[Mapping[str, slice]] = None,
            padding: Optional[bool] = False,
            **kwargs) -> np.ma.MaskedArray:
        """
        Return the field values as a :class:`np.ma.MaskedArray` object.

        Args:
            index: any kind of xarray indexing compatible with xarray
                :func:`~xarray.DataArray.isel` selection method.

            padding: pad the result with fill values where slices are out of the
             field dimension limits. Default is False.

        Returns:
            The field values as a np MaskedArray

        """
        return self._array.cerbere.isel(
            index, padding=padding, as_masked_array=True, **kwargs)

    def get_xvalues(
            self,
            index: Optional[Mapping[str, slice]] = None,
            padding: Optional[bool] = False,
            **kwargs) -> xr.DataArray:
        """
        Return the field values as a :class:`xarray.DataArray` object.

        Args:
            index: any kind of xarray indexing compatible with
                :func:`xarray.DataArray.isel` selection method.

            padding: pad the result with fill values where slices are out of the
             field dimension limits. Default is False.

        Returns:
            The field values as a xarray DataArray

        """
        return self._array.cerbere.isel(
            index, padding=padding, as_masked_array=False, **kwargs)

    def set_values(
            self,
            values: Union[np.ndarray, np.ma.MaskedArray],
            index: Optional[Mapping[str, slice]] = None,
            **kwargs) -> None:
        """set the values of a field.

        It is  possible to set only a subset of the field data array, using
        ``index``:

        >>> import numpy as np
        >>> data = np.ma.zeros((100, 200))
        >>> field = Field(data, name='test', dims=('x', 'y'))
        >>> field.set_values(
        >>>        np.full((10, 5,), 2),
        >>>        {'x': slice(10, 20), 'y': slice(0, 5)}
        >>>        )

        Args:
            values: the values to replace the ones in the field
            index: a dict of slices or indices of the subset to replace in the
                current field data array

        """
        if index is None:
            self._array[:] = values
        else:
            self._array.loc[index] = values

    def is_saved(self):
        """
        Return True is the content of the field is saved on file and
        was not updated since
        """
        raise DeprecationWarning

    def bitmask_or(
            self,
            meanings,
            index: Mapping[Hashable, Any]=None,
            **kwargs):
        """helper function to test specific bits in a CF compliant mask

        Bit (or flag) fields are arrays of integers where each bit has a
        specific meaning, described in a ``flag_meaning`` field attribute.
        Providing a list of the meanings to be tested in ``meaning``, a boolean
        mask if built, using the `or` logical operator, with a True value
        everywhere at least one of the provided meanings is set.

        The field must defined as in CF convention, with ``flags_masks``
        and ``flag_meanings`` field attributes.

        Args:
            meanings(list or str): a list of the meanings that have to be set
                or a str if only one bit is tested. Available meanings are
                listed in the ``flag_meanings`` attribute of the field.

        Returns:
            A boolean array.
        """
        return self._array.cerbere.isel(index, **kwargs).cerbere.bitmask_or(
            meanings)

    @classmethod
    def compute(
            cls,
            func,
            field1: 'Field', field2: 'Field'=None,
            **kwargs) -> 'Field':
        """Apply a function to a field (possibly combining with a second one)
         and returns the result as a new field.

        The function may be for instance a np MaskedArray operator
        such as np.ma.anom, np.ma.corr,...

        To be used with caution.

        Args:
            func (function) : the function to be called (ex: np.ma.anom)
            field1 (Field) : the field argument to the function
            field2 (Field, optional) : an optional 2nd field argument to the
                function
            kwargs : any argument to Field creation further describing the
                returned Field (units, name, ...).

        Returns:
            Field: the result field
        """
        if 'name' not in kwargs:
            varname = 'result'
        if field2 is None:
            values = func(field1.get_values())
        else:
            values = func(field1.get_values(), field2.get_values())
        field = Field(data=values,
                      name=varname,
                      dims=copy.copy(field1.dims),
                      dtype=field1.dtype,
                      fill_value=field1.fill_value,
                      **kwargs)
        return field

    def clone(
            self,
            index: Mapping[Hashable, Any] = None,
            padding: bool = False,
            prefix: str = None,
            **kwargs) -> 'Field':
        """Create a copy of a field, or a subset defined by index, and
        padding out as required.

        The data are also copied.

        The returned field does not contain any attachment to the source file
        attached to the original field, if any.

        Args:
            index (dict, optional):any kind of xarray indexing compatible with
                xarray :func:`~xarray.DataArray.isel` selection method.
            padding (bool, optional): True to pad out feature with fill values
                to the extent of the dimensions.
            prefix (str, optional): add a prefix string to the field names of
                the extracted subset.
        """
        if index is None:
            new_field = Field(
                data=self._array.copy(deep=True)
            )
        else:
            new_index = {
                dim: val
                for dim, val in index.items() if dim in self._array.dims
            }
            subarray = self._array[new_index]
            if padding:
                data = self.__pad_data(subarray.values, new_index)
                subarray.set_values(data)

            new_field = Field(data=subarray.copy(deep=True))

        # # copy cerbere internal encoding attributes...
        # new_field.encoding['cerbere'] = copy.copy(
        #     self._array.encoding['cerbere'])
        #
        # # ...but detach from any parent dataset
        # self.detach()

        if prefix is not None:
            new_field.set_name(prefix + new_field.name)
        return new_field

    def detach(self):
        self.dataset = None

    def rename(self, newname: str) -> None:
        """Rename the field inplace.

        Args:
            newname (str): new name of the field
        """
        self.name = newname

    def __add__(self, other: 'Field') -> 'Field':
        """Return a new field with the sum of current and an other field."""
        res = Field.convert_from_xarray(self.xrdata + other.xrdata)
        res.xrdata.name = '{}_{}_sum'.format(self.name, other.name)
        return res

    def __sub__(self, other: 'Field') -> 'Field':
        """Return a new field with the difference of current and an other
        field.
        """
        res = Field.convert_from_xarray(self.xrdata - other.xrdata)
        res.xrdata.name = '{}_{}_difference'.format(self.name, other.name)
        return res

    def module(self) -> 'Field':
        """Return the module field from its eastward and northward components.

        The module is sqrt(u² + v²)

        Returns:
            Field: the module field
        """
        u, v = self.components
        values = np.ma.sqrt(
            np.ma.power(u.get_values(), 2) +
            np.ma.power(v.get_values(), 2)
        )
        if 'eastward_' in u.name:
            modname = u.name.replace('eastward_', '')
        else:
            modname = 'module'
        field = Field(
            values,
            name=modname,
            dims=u.dims,
            dtype=u.dtype,
            fill_value=u.fill_value,
            units=u.units
        )
        return field



# -*- coding: utf-8 -*-
"""
Feature class for the trajectory observation pattern.
"""
from __future__ import absolute_import, division, print_function

from cerbere.feature.feature import Feature

__all__ = ['Trajectory']


class Trajectory(Feature):
    """
    Feature class for the trajectory observation patterns.

    A :class:`~cerbere.feature.trajectory.Trajectory` feature is defined by
    geolocation coordinates field: ``lon(time)``, ``lat(time)``,
    ``time(time)``, optionally ``z(time)``, all having a single geolocation
    dimension: ``time``.
    """
    _instance_dimname = 'trajectory'
    _feature_geodimnames = 'time',

    def get_geocoord_dimnames(self,  *args, **kwargs):
        return 'time',



# -*- coding: utf-8 -*-
"""
Feature class for orthogonal multidimensional collection of features.
"""
import typing as T

import xarray as xr

import cerbere
from cerbere.feature.cbasecollection import BaseCollection
from cerbere.feature.cbasefeature import BaseFeature


__all__ = ['OMDCollection']


class OMDCollection(BaseCollection):
    """
    Feature class for an orthogonal multidimensional collection of features.

    A :class:`~cerbere.feature.omdcollection.OMDCollection` class is defined by
    geolocation coordinates field and geolocation dimension of the feature, plus
    the instance dimension defined for this feature.

    For instance, for a :class:`~cerbere.feature.profile.Profile` feature, the
    instance dimension is ``profile``; the CF convention states that the
    geolocation coordinates field should be defined as: ``lon(profile)``,
    ``lat(profile)``, ``time(profile)``, ``z(z)``; observation field should be
    defined as ``fieldname(profile, z)``.
    """
    @classmethod
    def from_list(
            cls,
            items: T.List[BaseFeature],
            feature_class: BaseFeature = None) -> 'OMDCollection':
        """Build a collection feature from a list of feature instances"""
        if feature_class is None:
            feature_class = items[0].__class__

        dst = xr.concat(
            [_.dataset for _ in items], dim=items[0]._instance_dimname
        ).rename_dims(
            {items[0]._instance_dimname: feature_class._instance_dimname})

        return OMDCollection(
            dst,
            feature_class=feature_class
        )

    @classmethod
    def _canonical_dims(
            cls, coord: str, instance_class=None
    ) -> T.Union[T.List[T.Set[str]], T.Set[str]]:
        if coord == 'vertical':
            if isinstance(instance_class.cf_canonical_dims[coord], list):
                return [set(_) - {instance_class.cf_instance_axis}
                        for _ in instance_class.cf_canonical_dims[coord]]
            else:
                return set(instance_class.cf_canonical_dims[coord]) - \
                    {instance_class.cf_instance_axis}
        else:
            return set(instance_class.cf_canonical_dims[coord])

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset):

        for instance_class in ['Profile', 'Point']:
            try:
                cls._check_feature(dataset, instance_class=instance_class)
                print(instance_class)
                return cls
            except cerbere.AxisError:
                pass

        # check if there is an instance variable
        instance_dims = dataset.cb.cf_instance_dims

        if len(instance_dims) == 0:
            return

        # to be a collection, the instance_var should have a dimension of
        # the same name
        for i_dim in instance_dims:
            if i_dim not in dataset.dims:
                return

        instance_dim = dataset.cb.cf_instance_dims[0]

        # verify conformity to OMD Collection
        # coordinates must be: X(i), Y(i), T(i), Z(z)
        inst_size = dataset.sizes[instance_dim]

        # lon(instance_dim,...)
        check_lon = cls._check_coord_dims(
            dataset, 'lon', [instance_dim], inst_size)
        # lat(instance_dim,...)
        check_lat = cls._check_coord_dims(
            dataset, 'lat', [instance_dim], inst_size)
        # time(instance_dim,...)
        check_time = cls._check_coord_dims(
            dataset, 'time', [instance_dim], inst_size)

        check_z = True
        # two cases
        # z(z)
        # z() : degenerate case with no Z axis
        if dataset.cb.vertical is not None and 'Z' in dataset.cb.cf_axis_dims:
            z_size = dataset.cb.vertical.sizes

            # z(z)
            check_z = len(z_size) == 1 and cls._check_coord_dims(
                dataset, dataset.cb.vertical.name,
                [dataset.cb.vertical.name],
                list(z_size.values())[0])

        if all([check_lon, check_lat, check_time, check_z]):
            return cls




"""
A collection of functions to harmonize Xarray Dataset and DataArray objects
wrt to CF and other conventions
"""
import datetime
import dateutil.parser as parser
from itertools import chain
import logging
import typing as T
import warnings

import numpy as np
import xarray as xr

import cerbere
import cerbere.cfconvention as cf
import cerbere.accessor.matching as match
from cerbere.feature.cdiscretefeature import DiscreteFeature


# key of Dataset encoding attribute for identified CF Coordinates
CB_COORDS = '_cb_coords'

# key of Dataset encoding attribute for dimensions being axes
CB_AXIS_DIMS = '_cb_axis_dims'

#
CB_AXIS_COORDS = '_cb_axis_coords'

# key of Dataset encoding attribute for dimensions being instance axes in a
# discrete feature collection
CB_INSTANCE_DIMS = '_cb_instance_dims'

# key of Dataset encoding attribute for coordinates being instance axes in a
# discrete feature collection
CB_INSTANCE_COORDS = '_cb_instance_coords'

# key of encoding attribute set when Dataset object was harmonized with cerbere
CB_CERBERIZED = '_cb_is_cerberized'


def _cerberize(
        dst: xr.Dataset,
        dim_matching: T.Optional[T.Dict[str, str]] = None,
        coord_matching: T.Optional[T.Dict[str, str]] = None,
        attr_matching: T.Optional[T.Dict[str, str]] = None,
        axis_coordinates: T.Optional[T.Dict[str, str]] = None) -> xr.Dataset:
    """apply some transformation to original dataset to make it more
    convention compliant (CF 1.7+).

    Args:
        coord_matching (dict): explicitly provides the matching between file
            native coordinate vars (keys) and their expected Cerbere/CF
            equivalent (values).
        axis_coordinates(dict): explicitly provide the matching between
          CF spatiotemporal axis dimensions (X, Y, Z, T) and the
          coordinate variables in the dataset.
        attr_matching (dict): explicitly provides the matching between file
            native attributes (keys) and their expected Cerbere/CF
            equivalent (values).
    """
    # first modify the dataset according to the method arguments
    if dim_matching is not None:
        dst = dst.swap_dims(dim_matching)

    if coord_matching is not None:
        dst = dst.rename(coord_matching)

    if axis_coordinates is not None:
        for axis in axis_coordinates:
            dst[axis_coordinates[axis]].cb.axis = axis

    if attr_matching is not None:
        for attr, new_attr in attr_matching.items():
            dst.attrs[new_attr] = attr

    # store found CF axis dimensions
    if CB_AXIS_DIMS not in dst.encoding:
        dst.encoding[CB_AXIS_DIMS] = {}
    if CB_AXIS_COORDS not in dst.encoding:
        dst.encoding[CB_AXIS_COORDS] = {}
    if CB_INSTANCE_DIMS not in dst.encoding:
        dst.encoding[CB_INSTANCE_DIMS] = []
    if CB_INSTANCE_COORDS not in dst.encoding:
        dst.encoding[CB_INSTANCE_COORDS] = []

    # verify and make compliant spatiotemporal coordinates and dimensions
    dst = _cerberize_coords(dst)
    dst = _cerberize_dims(dst)

    # guess discrete instance dim
    # detect in the dimensions of CF coordinates the ones that don't
    # correspond to spatiotemporal dimensions
    # instance_dim = {}
    # for c in cf.GEOCOORDS:
    #     coord = self._cfdataset[c]
    #     if coord is None:
    #         assert c == 'vertical'
    #         continue
    #
    #     instance_dims = set(coord.dims) - {'time', self.Z, 'lat', 'lon',
    #                                        'cell', 'row', 'ni', 'nj'}
    #     if len(instance_dims) != 0:
    #         instance_dim = instance_dim.union(instance_dims)
    #     # @TODO

    # guess coordinate axis
    # self._cfdataset = self._cfdataset.cf.guess_coord_axis(verbose=False)

    # set lat, lon, time as coordinates and complete metadata to match CF
    # model
    for coord in dst.encoding[CB_COORDS]:

        # ensure it is set as coordinate
        if coord in dst.variables:
            if coord not in dst.coords:
                dst = dst.set_coords([coord])
            coordvar = dst.coords[coord]

        # axis coordinate => set axis attribute
        reverse_axis_dims = {
            v: k for k, v in dst.encoding[CB_AXIS_DIMS].items()}
        if coord in reverse_axis_dims:
            dst[coord].attrs['axis'] = reverse_axis_dims[coord]

        # if coordinate has no spatiotemporal dimensions and there is no
        # other variable corresponding to the same axis, then it must be an
        # axis coordinate
        # if len(coord.dims) == 0:
        #     # @TODO check CF convention about this
        #     if ('axis' not in coordvar.attrs
        #             or coordvar.attrs['axis'] not in cf.CF_COORD_AXIS.values()):
        #         coordvar.attrs['axis'] = cf.CF_COORD_AXIS[coord]

    # verify _FillValue and standardize
    for v in dst.variables:
        # replace missing_value with _FillValue
        if 'missing_value' in dst.variables[v].encoding:
            if '_FillValue' in dst.variables[v].encoding:
                dst.variables[v].encoding.pop('missing_value')
            else:
                dst.variables[v].encoding['_FillValue'] = \
                    dst.variables[v].encoding.pop('missing_value')

    # harmonize attributes
    dst = _cerberize_attrs(dst)

    # update coordinates attribute of data variables
    dst = update_cf_coordinates_attr(dst)

    # check geolocation coordinates are defined and valid
    if not _check_mandatory_geocoordinates(dst):
        raise ValueError(
            'The dataset is not a valid observation set or is malformed.'
        )

    # mark as cerberized
    dst.encoding[CB_CERBERIZED] = True

    return dst


def _check_mandatory_geocoordinates(dst: xr.Dataset) -> bool:
    """check if the required CF spatiotemporal coordinates are in the
    dataset
    """
    ok = True
    for coord in cf.CF_REQ_GEOCOORDS:
        try:
            _ = _guess_cf_coord(dst, coord)
        except cerbere.CoordinateError as e:
            logging.exception(e)
            logging.warning(
                '  => Missing coordinate var: {}'.format(coord))
            ok = False

    return ok


def _cerberize_dims(dst: xr.Dataset) -> xr.Dataset:
    """Guess and harmonize the spatio-temporal axis dimensions
    """
    cf_axis_dims, cf_axis_coords, cf_instance_dims, = _guess_cf_axes(dst)

    # rename axis dimension to coordinate name if unidimensional, and set
    # axis attribute
    for axis, coord in cf_axis_coords.items():
        if (len(dst[coord].dims) == 1 and
                len(cf_axis_dims) > 0 and
                axis in cf_axis_dims and
                cf_axis_dims[axis] != coord):
            if cf_axis_dims[axis] in cf_instance_dims:
                # TODO do sth clearer here
                continue
            logging.debug(
                f'Renaming axis dimension {axis} to {coord} as it is an '
                f'axis coordinate')
            # dst = dst.swap_dims({cf_axis_dims[axis]: coord})
            # bug issue on swap_dims, replace above with rename :
            # https://github.com/pydata/xarray/issues/9542#issuecomment
            # -2372117209
            dst = dst.rename({cf_axis_dims[axis]: coord})
            cf_axis_dims[axis] = coord

        # if dst[coord].cb.axis is None:
        #     dst.cb.set_axis(coord, cf_axis_dims[axis])

    # rename instance dimension if any
    if len(cf_instance_dims) > 1:
        raise NotImplementedError

    if len(cf_instance_dims) == 1:
        ftype = dst.attrs.get('featureType')
        if ftype is not None:
            feat_class = cerbere.class_by_cf_feature_type(ftype)
            if feat_class.cf_element_instance_axis is None:
                i_axis = feat_class.cf_instance_axis
            else:
                i_axis = feat_class.cf_element_instance_axis
            if i_axis is not None and cf_instance_dims[0] != i_axis:
                logging.debug(f'Renaming instance axis {cf_instance_dims[0]} '
                              f'to {i_axis}')
                dst = dst.swap_dims({cf_instance_dims[0]: i_axis})
        else:
            i_axis = cf_instance_dims[0]

        dst.encoding[CB_INSTANCE_DIMS] = (i_axis,)

    # name = None
    # for cfname in cf.CF_AXIS:
    #     try:
    #         name = self._guess_cf_axis(self._cfdataset, cfname)
    #     except CoordinateError:
    #         if cfname == 'vertical':
    #             logging.debug('No coordinate found for vertical axis')
    #         else:
    #             raise

    reduced = {}
    for dim in dst.dims:
        if dim in match.DIM_MATCHING:
            if dim == match.DIM_MATCHING[dim]:
                continue
            if match.DIM_MATCHING[dim] in dst.dims:
                if match.DIM_MATCHING[dim] in match.DIM_MATCHING:
                    continue
                raise ValueError(
                    'Attempt to rename a dimension with the same name as an'
                    ' already existing dimension, without permutation name '
                    'provided : {} <=> {}'.format(dim, match.DIM_MATCHING[dim]))
            reduced[dim] = match.DIM_MATCHING[dim]

    dst.encoding[CB_AXIS_DIMS] = cf_axis_dims
    dst.encoding[CB_AXIS_COORDS] = cf_axis_coords

    return dst
#    return dst.swap_dims(reduced)


def _cerberize_coords(dst) -> xr.Dataset:
    """
    Guess and harmonize the spatio-temporal coordinate variables
    """
    # store found coordinates
    if '_cb_coords' not in dst.encoding:
        dst.encoding[CB_COORDS] = {}

    name = None
    for cfname in cf.CF_COORDS:
        try:
            name = _guess_cf_coord(dst, cfname)
        except cerbere.CoordinateError:
            if cfname == 'vertical':
                logging.debug('No coordinate found for vertical axis')
                continue
            else:
                raise

        # rename latitude/longitude/time to Cerbere convention: lat, lon, time
        # z coordinate is not renamed but accessible through property `vertical`
        cb_name = cf.CB_COORD_CONVENTION.get(cfname, name)
        if cb_name != name:
            dst = dst.rename_vars({name: cb_name})

        # fill in missing attributes
        if cfname in cf.GEOCOORD_ATTRS:
            for attr, val in cf.GEOCOORD_ATTRS[cfname].items():
                if attr not in dst[cb_name].attrs:
                    dst[cb_name].attrs[attr] = val

        # memorize found coordinates
        dst.encoding[CB_COORDS][cfname] = cb_name

        # set as a dataset coordinate
        if cb_name not in dst.coords:
            dst = dst.set_coords(cb_name)

    # check all mandatory coordinates found
    for coord in cf.CF_REQ_GEOCOORDS:
        if coord not in dst.encoding[CB_COORDS]:
            raise cerbere.CoordinateError(
                f'Mandatory coordinate {coord} was not found in this '
                f'dataset')

    # # rename variables wrt to a dictionary of aliases. By default the
    # # aliases correspond to frequently met variant of spatial and
    # # temporal coordinate names
    # reduced = {}
    # for v in self._cfdataset.variables:
    #     if v in var_matching:
    #         if v == var_matching[v]:
    #             continue
    #         if var_matching[v] in self._cfdataset.variables:
    #             if var_matching[v] in var_matching:
    #                 continue
    #             raise ValueError(
    #                 'Attempt to rename a variable with the same name as an'
    #                 ' already existing variable, without permutation name '
    #                 'provided : {} <=> {}'.format(v, var_matching[v]))
    #
    #         reduced[v] = var_matching[v]
    #
    # self._cfdataset = self._cfdataset.rename_vars(reduced)

    # # guess axis coordinates
    # if self._cfdataset.encoding.get('cf_axes') is None:
    #     self._cfdataset.encoding['cf_axes'] = (
    #         DatasetAccessor._guess_cf_axes(self._cfdataset))
    #
    # # find spatial and temporal coordinate variables
    # for axis in ['X', 'Y', 'T', 'Z']:
    #     if axis in self._cfdataset.data_vars and \
    #             len(self._cfdataset.data_vars[axis]) == 1:
    #         pass

    return dst


def _cerberize_attrs(
        dst: xr.Dataset,
        attr_matching: T.Optional[T.Dict[str, str]] = None):
    """decode the standard attributes

    Possibly change the name of some attributes to harmonize them to
    Cerbere CF/ACDD convention, based on an internal dict of known variants.
    """
    if attr_matching is not None:
        for stdattr, attr in attr_matching.items():
            dst.attrs[stdattr] = dst.attrs.pop(attr)

    _cerberize_time_coverage(dst)

    return dst


def _cerberize_time_coverage(dst: xr.Dataset):
    """decode time coverage attributes and update the translation table"""
    try:
        dt = _decode_time_coverage_attr(dst, 'start')
        if dt is not None:
            dst.attrs['time_coverage_start'] = dt
        dt = _decode_time_coverage_attr(dst, 'end')
        if dt is not None:
            dst.attrs['time_coverage_end'] = dt
    except (TypeError, ValueError) as e:
        logging.error(f'error when trying to get the time coverage from '
                      f'attributes; {str(e)}')
    except:
        raise


def _decode_time_coverage_attr(
        dst: xr.Dataset, boundary: str) -> datetime.datetime:
    """Convert start/end time attributes to Cerbere/CF convention (
    time_coverage_start/time_coverage_end), attempting different existing
    conventions.

    Args:
        boundary(str): type of date attribute (`start` or `end`)

    Returns:
        datetime: start/end time of the data in file.
    """
    attrs = dst.attrs

    # time_coverage_start/end
    for attrstyle in match.TIME_COVERAGE_ATTRS:
        attrdate = match.TIME_COVERAGE_ATTRS[attrstyle][boundary]
        if attrdate in attrs:
            if isinstance(attrs[attrdate], datetime.datetime):
                return attrs.pop(attrdate)
            return parser.parse(attrs.pop(attrdate))

    # combination of two time/date attributes (such as start_date and
    # start_time)
    attrdate = match.TIME_COVERAGE_ATTRS['date'][boundary]
    if attrdate in attrs:
        if match.TIME_COVERAGE_ATTRS['time'][boundary] in attrs:
            # start/end_time/date combination
            attrtime = attrs.pop(
                match.TIME_COVERAGE_ATTRS['time'][boundary])
            return parser.parse(f'{attrs.pop(attrdate)}T{attrtime}')

        return parser.parse(attrs.pop(attrdate))


def update_cf_coordinates_attr(dst: xr.Dataset) -> xr.Dataset:
    """Update the "coordinates" attribute of data variable according to CF
    rules.
    """
    aux_coords = guess_cf_aux_coords(dst)

    if len(aux_coords) > 0:
        for v in dst.data_vars:
            coordinates = []
            for aux, adims in aux_coords.items():
                if len(set(dst[v].dims).intersection(set(adims))) > 0:
                    coordinates.append(aux)
            if len(coordinates) > 0:
                if ('coordinates' in dst[v].encoding
                        and set(dst[v].encoding['coordinates'].split(' ')) !=
                        set(coordinates)):
                    warnings.warn(
                        f'The current value of "coordinates" attribute of '
                        f'variable {v} does not equal what was inferred by '
                        f'cerbere: '
                        f'{set(dst[v].encoding["coordinates"].split(" "))} <> '
                        f'{set(coordinates)}')
                    continue
                dst[v].encoding['coordinates'] = ' '.join(coordinates)

    return dst


def guess_cf_aux_coords(dst: xr.Dataset) -> dict[T.Any, tuple[T.Hashable, ...]]:
    """Guess the spatiotemporal coordinates of a dataset, among longitude,
    latitude, vertical or time, which depends on more than one spatiotemporal
    dimension"""
    aux_coords = {}
    for coord in dst.encoding[CB_COORDS].values():
        if len(dst[coord].dims) > 1:
            aux_coords[coord] = dst[coord].dims

    return aux_coords


def _guess_cf_coord_on_units(
        dst: xr.Dataset,
        name: str) -> T.Set[T.Hashable]:
    if name == 'time':
        return {
            _ for _ in dst.variables
            if ('units' in dst[_].attrs and dst[_].attrs['units'] is str and
               dst[_].attrs['units'].startswith('since '))
               or np.issubdtype(dst[_].dtype, np.datetime64)}

    if name not in cf.CF_COORD_UNITS:
        # currently difficult to identify Z axis as many units possible
        # which can be confused with other quantities
        return set()

    return {
        _ for _ in dst.variables
        if 'units' in dst[_].attrs and
           dst[_].attrs['units'] in cf.CF_COORD_UNITS[name]}


def _guess_cf_coord_on_std_name(
        dst: xr.Dataset,
        name: str) -> T.Set[T.Hashable]:
    cf_std_name = cf.GEOCOORD_STD_NAME[name]
    matches = {
        _ for _ in dst.variables.keys()
        if dst[_].attrs.get('standard_name') in cf_std_name}
    if len(matches) > 1:
        logging.warning(
            f'There are more than one variable with a standard name '
            f'matching a possible {name} coordinate. The coordinate for '
            f'this axis may not be identified and '
            f'it will probably require a specific reader to fix this '
            f'dataset.')

    return matches


def _guess_cf_coord_on_alias(
        dst: xr.Dataset,
        name: str) -> T.Set[T.Hashable]:
    # look for aliases
    return {
        _ for _ in dst.variables.keys() if _ in match.COORD_ALIASES[name]}


def _guess_axis_coord(dst: xr.Dataset, axis: str) -> T.List[T.Hashable]:
    """Guess the coordinate variable corresponding to a given axis attribute
    in a Dataset

    Args:
        dst: Dataset
        axis: CF coordinate axis name (X, Y, Z, T)

    Return:
        the list of variable names matching the coordinate axis in the Dataset
    """
    coords = [
        _ for _ in dst.variables.keys() if dst[_].attrs.get('axis') == axis]
    if len(coords) > 1:
        warnings.warn(
            f'There are more than one variable with attribute axis'
            f' {axis}: {coords}. There must be only one in a CF compliant '
            f'file. The coordinate for this axis may not be identified and '
            f'it will probably require a specific reader to fix this '
            f'dataset.', cerbere.CWarning)

    return coords


def _guess_cf_coord_on_axis(
        dst: xr.Dataset,
        name: str) -> T.List[T.Hashable]:
    """Guess geocoordinate (latitude, longitude, time or vertical) variable
    based on axis attribute"""
    return _guess_axis_coord(dst, cf.CF_COORD_AXIS[name])


def _guess_cf_coord(
        dst: xr.Dataset,
        name: str) -> T.Hashable:
    """.
    Find in the dataset the corresponding coordinate name for a CF
    standard spatiotemporal coordinate.

    The search is based on, in order of priority:
    * the units of the dataset variables
    * the standard name attribute of the dataset variables
    * the matching with a list of common aliases for the CF
      spatiotemporal coordinates
    """
    # search based on units
    units = _guess_cf_coord_on_units(dst, name)

    # search a variable with a matching standard name for the given
    # coordinate
    std_attr = _guess_cf_coord_on_std_name(dst, name)

    # look for aliases
    aliases = _guess_cf_coord_on_alias(dst, name)

    # look for axis variables
    axes = set(_guess_cf_coord_on_axis(dst, name))

    # attempt different selection strategies
    candidates = units.union(std_attr, aliases, axes)

    if len(candidates) == 0:
        if name in dst.variables:
            return name
        raise cerbere.CoordinateError(
            f'No candidate variable for coordinate {name}')

    if len(candidates) == 1:
        # agreement of all criteria
        return candidates.pop()

    for cname in [name, cf.CB_COORD_CONVENTION.get(name)]:
        if cname in dst.variables and cname in candidates:
            return cname

    # repeated unique candidate ?
    repeated = candidates
    for _ in [units, std_attr, aliases, axes]:
        if len(_) != 0:
            repeated = repeated.intersection(_)
    if len(repeated) == 1:
        return repeated.pop()

    # ambiguous case
    logging.warning(
        f'several candidate variables for coordinate {name}: '
        f'{repeated}.')

    if len(axes) == 1:
        # precedence of axis information
        logging.warning(f'selecting {list(axes)[0]} as coordinate {name} based '
                        f'on axis indication')
        return list(axes)[0]

    if len(units.union(std_attr)) == 1:
        logging.warning(f'selecting {list(units.union(std_attr))[0]} as '
                        f'coordinate {name} based on units and/or standard '
                        f'attribute indication')
        return units.union(std_attr).pop()

    raise cerbere.CoordinateError(
        f'several candidate variables for coordinate {name}: {aliases}')

#
# def get_geocoord_by_axis(dst: xr.Dataset, axis: str) -> xr.DataArray:
#     """Return the geolocation coordinate variable for the given axis.
#     """
#     axis = dst.coords[dst.cb.axis_coordname(axis)]
#
#     if axis is None:
#         raise cerbere.AxisError(f'No coordinate for {axis} axis')
#
#     return axis
#
#
# def axis_coordname(dst: xr.Dataset, axis: cf.CF_AXIS) -> str:
#     """Returns the name of the spatiotemporal coordinate associated
#     with a given CF axis.
#
#     Args:
#         axis: axis name among 'X', 'Y', 'T', 'Z'
#     """
#     if axis not in cf.CF_AXIS:
#         raise cerbere.AxisError(
#             '{} is not a geolocation coordinate'.format(axis))
#
#     for coordname in self.geocoordnames:
#         if self._cfdataset.coords[coordname].attrs['axis'] == axis:
#             if len(self._cfdataset.coords[coordname].dims) > 1:
#                 logging.warning(
#                     f'The variable {coordname} has an axis attribute with'
#                     f' value {axis} but it can not be an axis as it has '
#                     f'more than one dimension. The file or Dataset object '
#                     f'should be fixed first.')
#             else:
#                 return coordname
#
#     raise cerbere.AxisError('No coordinate found for axis: {}'.format(axis))


def _set_axis(dst: xr.Dataset, name: str, axis: str):
    """Set a variable as a CF axis.

    Args:
        name: name of the variable to set as an axis
        axis: value of the axis
    """
    if axis not in cf.CF_AXIS:
        raise cerbere.AxisError(f'Illegal axis {axis}')

    # verify there is no variable with this axis already
    axis_vars = set(
        [dst[_].name for _ in dst.variables
         if 'axis' in dst[_].attrs and dst[_].attrs['axis'] == axis]) - {name}
    if len(axis_vars) > 0:
        raise cerbere.AxisError(
            f'Axis {axis} is already set for variable(s) {axis_vars}. '
            f'Only one coordinate variable is possible for a given axis.')

    dst[name].attrs['axis'] = axis


def _guess_cf_instance_coord(dst: xr.Dataset) -> T.Optional[T.Hashable]:
    """the name of the instance variable, holding the `cf_role`
    attribute

    One-dimensional variables in a Discrete Geometry CF file, which have
    only the instance dimension as dimension (such as x(i), y(i) and z(i)
    for a timeseries), are instance variables. Instance variables provide
    the metadata that differentiates individual features.
    """
    cf_role_vars = [v for v in dst.variables if 'cf_role' in dst[v].attrs]

    if len(cf_role_vars) == 1:
        return cf_role_vars[0]

    if len(cf_role_vars) == 2:
        # possible for
        # @TODO
        pass

    if len(cf_role_vars) > 2:
        raise cerbere.CoordinateError(
            f'There are multiple variables with `cf_role` attribute. It '
            f'seems like a non-compliance'
        )

    dst.encoding[CB_INSTANCE_COORDS] = cf_role_vars


def _guess_cf_instance_dim(dst: xr.Dataset) -> T.Optional[T.Hashable]:
    """the instance dimension of the dataset, if any.

    The instance dimension is the dimension that identifies a particular
    feature within a collection of features in CF convention.

    Instance dimensions may be provided for CF data models such as:
      - Profile
      - Collections
    """
    inst_var = _guess_cf_instance_coord(dst)

    # instance variable is defined
    if inst_var is not None:
        if len(dst[inst_var].dims) > 1:
            raise ValueError('A CF instance variable should have only '
                             'zero or one dimension')
        if len(dst[inst_var].dims) == 1:
            return dst[inst_var].dims[0]

    # check among instance dim names known in CF convention for
    # different feature types if one is present in the dataset dims
    # (profile, station, trajectory,...)
    for feat_class in cerbere.helpers._feature.values():
        if not isinstance(feat_class, DiscreteFeature):
            return
        axis = feat_class.load().cf_instance_axis
        if axis in dst.dims:
            return axis


def _guess_explicit_axis_coordinates(
        dst: xr.Dataset) -> T.Tuple[T.Dict, T.Dict]:
    """guess and return the axis coordinates with explicit axis attribute

    Return:
        a tuple of two dictionaries where keys are the CF axes (X, Y, Z or T)
        and values, respectively, the corresponding axis coordinate variable
        names and axis coordinate dimensions.
    """
    cf_axis_coords = {}
    cf_axis_dims = {}

    for axis in cf.GEOCOORD_AXIS.values():
        axis_attr_coords = _guess_axis_coord(dst, axis)
        if len(axis_attr_coords) != 1:
            # ambiguous: more than one variable with same axis attribute value
            continue

        axis_attr_coord = axis_attr_coords[0]
        if len(dst[axis_attr_coord].dims) != 1:
            # not an axis coordinate
            continue

        axis_coord_dim = dst[axis_attr_coord].dims[0]
        cf_dims = dict(zip(*np.unique(
            list(chain.from_iterable(dst[_].dims for _ in dst.coords)),
            return_counts=True)))
        if cf_dims[axis_coord_dim] == 1:
            # unique coordinate variable depending on this axis
            cf_axis_coords[axis] = axis_attr_coord
            cf_axis_dims[axis] = axis_coord_dim
        elif axis_coord_dim == axis_attr_coord:
            # coordinate and dimension share same name
            cf_axis_coords[axis] = axis_attr_coord
            cf_axis_dims[axis] = axis_coord_dim

    return cf_axis_coords, cf_axis_dims


def _guess_cf_axes(dst: xr.Dataset) -> T.Tuple[
        T.Dict[str, T.Hashable], T.Dict[str, str], T.List[T.Hashable]]:
    """Guess the CF axis dimensions in a dataset and return a dictionary
    where keys are the found CF axes and values the corresponding
    coordinate names in the dataset.
    """
    # check coordinates have been guessed already
    assert CB_COORDS in dst.encoding

    # dimensions that are CF standard axes (X, Y, Z, T)
    # key = standard axis, value = dimension name in dataset
    cf_axis_dims: T.Dict[str, T.Hashable] = {}

    # coordinates that are considered as axes as they depend on an axis
    # dimension not shared without other coordinates or are unidimensional
    cf_axis_coords: T.Dict[str, str] = {}

    # special axis corresponding to a feature instance in a discrete collection
    cf_instance_dims: T.Set[T.Hashable] = set()

    # get the dimension axes used by the CF coordinates and their
    # occurrences
    cf_dims = dict(zip(*np.unique(
        list(chain.from_iterable(
            dst[_].dims for _ in dst.encoding.get(CB_COORDS).values()
        )),
        return_counts=True)))

    # get axis coordinates with explicit axis attribute
    _cf_axis_coords, _cf_axis_dims = _guess_explicit_axis_coordinates(dst)
    cf_axis_coords.update(_cf_axis_coords)
    cf_axis_dims.update(_cf_axis_dims)

    # identify the axes unique to some coordinate = axis coordinates
    for cf_coord in cf.CF_COORDS:
        if cf_coord not in dst.encoding.get(CB_COORDS):
            # coordinate not existing or not identified in the dataset
            continue
        coord = dst.encoding.get(CB_COORDS)[cf_coord]

        # dimensionless coordinate must be an axis
        if len(dst[coord].dims) == 0:
            _set_axis(dst, coord, cf.GEOCOORD_AXIS[cf_coord])
            #cf_axis_coords[cf.GEOCOORD_AXIS[cf_coord]] = coord
            continue

        # latitude(dim_0), longitude(dim_1) => independant axes for lat , lon
        lat_dims = set(dst[dst.encoding.get(CB_COORDS)['latitude']].dims)
        lon_dims = set(dst[dst.encoding.get(CB_COORDS)['longitude']].dims)

        if cf_coord in ['latitude', 'longitude'] and \
                lat_dims.intersection(lon_dims) == set():
            cf_axis_coords[cf.GEOCOORD_AXIS[cf_coord]] = coord
            cf_axis_dims[cf.GEOCOORD_AXIS[cf_coord]] = dst[coord].dims[0]
            _set_axis(dst, coord, cf.GEOCOORD_AXIS[cf_coord])
            continue

        if len(dst[coord].dims) == 1:
            # coordinate has one unique axis
            udim = dst[coord].dims[0]

            # latitude(dim_0), longitude(dim_1)
            if cf_dims[udim] == 1:
                # this axis is only used by this coordinate => it is an
                # axis coordinate
                cf_axis_coords[cf.GEOCOORD_AXIS[cf_coord]] = coord
                cf_axis_dims[cf.GEOCOORD_AXIS[cf_coord]] = udim
                _set_axis(dst, coord, cf.GEOCOORD_AXIS[cf_coord])
                # @TODO Dirty workaround for TimeSeriesProfile. We need
                #  better identification/separation of instance and element
                #  variables
                if (len(lon_dims) == len(lat_dims) == 0 and
                        cf_coord != 'vertical' and
                        'Z' in cf_axis_coords):
                    cf_instance_dims = cf_instance_dims.union({udim})

            elif (cf_dims[udim] == len(dst.encoding.get(CB_COORDS)) or
                  ('vertical' in dst.encoding.get(CB_COORDS) and
                   cf_dims[udim] == len(dst.encoding.get(CB_COORDS))-1)):
                # all coords depend on the same dimension - if not time it
                # is guessed as an instance dimension
                if udim != 'time':
                    cf_instance_dims = {udim}
                elif cf.GEOCOORD_AXIS[cf_coord] == 'T':
                    cf_axis_dims[cf.GEOCOORD_AXIS[cf_coord]] = udim

                # coord is a dimensionless axis coordinate
                cf_axis_coords[cf.GEOCOORD_AXIS[cf_coord]] = coord
                _set_axis(dst, coord, cf.GEOCOORD_AXIS[cf_coord])

        elif len(dst[coord].dims) > 1:
            # the coordinate has multiple dimensions => mix of instance
            # and/or CF geospatial dimensions
            udim = [_ for _ in dst[coord].dims if cf_dims[_] == 1]
            if len(udim) == 1:
                # only one is unique to this coordinate => must be the
                # geospatial axis
                udim = udim[0]
                cf_axis_coords[cf.GEOCOORD_AXIS[cf_coord]] = coord
                cf_axis_dims[cf.GEOCOORD_AXIS[cf_coord]] = udim
                _set_axis(dst, coord, cf.GEOCOORD_AXIS[cf_coord])

                # remaining dim(s) must be collection instance dimension(s)
                cf_instance_dims = cf_instance_dims.union(
                    set(dst[coord].dims) - {udim}
                )

    # guess axis dimensions based on name
    for dim in dst.dims:
        if dim in match.DIM_MATCHING:
            axis_dim = match.DIM_MATCHING[dim]
            if (axis_dim in cf_axis_dims
                    and dim != cf_axis_dims[axis_dim]):
                warnings.warn(
                    f'Found more than one dimension for axis {axis_dim}: '
                    f'{dim} {cf_axis_dims[axis_dim]}. Keeping '
                    f'{cf_axis_dims[axis_dim]} as {axis_dim} axis',
                    cerbere.CWarning)
            else:
                cf_axis_dims[axis_dim] = dim

    dst.encoding[CB_AXIS_DIMS] = cf_axis_dims
    dst.encoding[CB_AXIS_COORDS] = cf_axis_coords

    return cf_axis_dims, cf_axis_coords, list(cf_instance_dims),

    # # find explicit axis coordinates having a proper CF variable
    # # attribute `axis`
    # missing_axis_coords = []
    # for axis in cf.CF_AXIS:
    #     try:
    #         cf_axis_coord[axis] = DatasetAccessor._find_axis_coord(dst, axis)
    #     except cerbere.AxisError:
    #         missing_axis_coords.append(axis)
    #
    #
    # if len(missing_axis_coords) == 0:
    #     # all axes found
    #     # verify they are truly axes
    #     DatasetAccessor._verify_cf_axes(dst, geo_axes)
    #
    #     # guess instance axes (for collection of features)
    #
    #     return geo_axes
    #
    # # check if spatiotemporal coordinates are axes
    # for coord in cf.CF_REQ_GEOCOORDS:
    #     dst_coord = dst.encoding[CB_COORDS].get(coord)
    #     axis = cf.GEOCOORD_AXIS[coord]
    #     if len(dst[dst_coord].dims) == 0 and axis not in geo_axes:
    #         geo_axes[axis] = dst_coord
    #
    # return geo_axes


# -*- coding: utf-8 -*-
"""
Feature class for the trajectory observation pattern.
"""
import typing as T

import xarray as xr

import cerbere
from cerbere.feature.cdiscretefeature import DiscreteFeature


__all__ = ['Trajectory']


class Trajectory(DiscreteFeature):
    """
    Feature class for the trajectory observation patterns.

    A :class:`~cerbere.feature.trajectory.Trajectory` feature is defined by
    geolocation coordinates field: ``lon(time)``, ``lat(time)``,
    ``time(time)``, optionally ``z(time)``, all having a single geolocation
    dimension: ``time``.
    """
    cf_instance_axis = 'trajectory'
    cf_feature_type = 'trajectory'
    cdm_data_type = 'trajectory'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'latitude': (cf_instance_axis, 'T'),
        'longitude': (cf_instance_axis, 'T'),
        'time': (cf_instance_axis, 'T'),
        'vertical': [(cf_instance_axis, 'T'), (cf_instance_axis,)],
        'data': (cf_instance_axis, 'T')
    }
    # optional axes for the feature
    cf_optional_axes = ['vertical']

    @classmethod
    def cerberize(cls, dataset: xr.Dataset, force_reorder: bool = True,
                  **kwargs):
        cfdataset = super(Trajectory, cls).cerberize(dataset, force_reorder)

        # CF specific attrs and vars
        # It is strongly recommended that there always be a trajectory
        # variable (of any data type) with the attribute
        # cf_role=”trajectory_id” attribute, whose values uniquely identify
        # the trajectories.
        # if cfdataset.cb.cf_role is None:
        #     cfdataset['trajectory'] = xr.DataArray(
        #         dims=(),
        #         attrs={'cf_role': 'trajectory_id'},
        #         data=0)

        # for _, v in cfdataset.data_vars.items():
        #     v.encoding['coordinates'] = 'time lat lon z'

        return cfdataset

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[DiscreteFeature]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if cls._unexpected_featureType(cfdst) or \
                cls._unexpected_cdm_data_type(cfdst):
            return

        try:
            cls._check_feature(dataset)
            return cls
        except cerbere.AxisError:
            pass

        # # check lat/lon/time are axes with a single time dimension
        # for axis in ['X', 'Y', 'T', 'Z']:
        #     if axis == 'Z' and axis not in dataset.cb.cf_axis_coords and \
        #             cfdst.cb.vertical is None:
        #         continue
        #     if not cls._check_coord_dims(
        #             dataset,
        #             dataset.cb.cf_axis_coords[axis],
        #             [dataset.cb.cf_axis_dims['T']]):
        #         return
        #
        # # check provided axes if any
        # # if cfdst.cb.X is not None and cfdst.cb.X.name != 'lon':
        # #     return
        # # if cfdst.cb.Y is not None and cfdst.cb.Y.name != 'lat':
        # #     return
        # # if cfdst.cb.T is not None and cfdst.cb.T.name != 'time':
        # #     return
        #
        # return Trajectory



"""
Test methods for a xarray/cerbere BackendEntryPoint to read a dataset.

All these methods must be imported in the test module for the new
BackendEntryPoint. This test module must implement the following fixtures:

* test_file : return the input file to be used in this test
* download_url : return the url to test data repository
* backend : return the BackendEntryPoint class name
* feature : return the related feature class corresponding to the dataset
"""
import datetime
import os
from pathlib import Path

import pytest

import numpy
import xarray as xr

import cerbere
from cerbere.feature.cbasefeature import BaseFeature


TEST_SUBSET = 'test_subset.nc'


@pytest.fixture(scope='module')
def download_url():
    return 'ftp://ftp.ifremer.fr/ifremer/cersat/projects/cerbere/test_data/'


@pytest.fixture(scope='module')
def input_file(test_file, backend, download_url):
    # search test file
    if 'CERBERE_TESTDATA_REPOSITORY' not in os.environ:
        raise EnvironmentError(
            'CERBERE_TESTDATA_REPOSITORY environment variable is not set')
    rootpath = Path(os.environ['CERBERE_TESTDATA_REPOSITORY'])
    dirname = Path(backend.split('.')[-1])
    testfilepath = rootpath / dirname / test_file

    if not testfilepath.exists():
        raise IOError(
            f'Test file {test_file} was not found. Please download it from '
              f'{download_url} and place in {Path(rootpath /dirname)} folder')

    return testfilepath


@pytest.fixture(scope='module')
def dataset(input_file, backend):
    """Open file and create dataset object"""
    return xr.open_dataset(input_file, engine=backend)


@pytest.fixture(scope='module')
def output_dir(tmpdir_factory):
    return tmpdir_factory.mktemp('output')


def test_open_file_as_dataset(dataset):
    """Test dataset instantiation with an input file"""
    assert isinstance(dataset, xr.Dataset)


def test_get_list_of_vars(dataset):
    """Test reading the list of fields"""
    for v in dataset.data_vars:
        assert isinstance(v, str)
        assert (v != "")


def test_read_vars(dataset):
    """Test reading and displaying variable properties"""
    vars = dataset.data_vars
    for v in vars:
        print(f'...reading {v}')
        assert isinstance(dataset.data_vars[v], xr.DataArray)
        print(dataset.data_vars[v])


def test_read_global_dimensions(dataset):
    """Test reading and displaying global dimensions"""
    print('...reading global dimensions')
    dims = dataset.dims

    assert dims is not None
    assert len(dims) > 0

    for dim in dims:
        dimsize = dataset.sizes[dim]
        print('.....', dim, dimsize)

        assert isinstance(dim, str)
        assert isinstance(dimsize, int)


def test_read_global_attributes(dataset):
    """Test reading the global attributes"""
    for attr in dataset.attrs:
        print(attr, dataset.attrs[attr])
        assert isinstance(dataset.attrs[attr],
                          (str, int, datetime.datetime, numpy.int32,
                           numpy.uint32, numpy.int16, numpy.float32,
                           numpy.float64, list))



from pathlib import Path
import re


import numpy as np
from xarray import open_dataset
from xarray.backends import BackendEntrypoint


class GHRSST(BackendEntrypoint):
    def open_dataset(
        self,
        filename_or_obj,
        mask_and_scale=True,
        decode_times=True,
        decode_timedelta=True,
        use_cftime=True,
        concat_characters=True,
        decode_coords=True,
        drop_variables=None,
        # other backend specific keyword arguments
        # `chunks` and `cache` DO NOT go here, they are handled by xarray
    ):
        ds = open_dataset(
            filename_or_obj, mask_and_scale=mask_and_scale,
            decode_times=decode_times, decode_timedelta=decode_timedelta,
            use_cftime=use_cftime, concat_characters=concat_characters,
            decode_coords=decode_coords, drop_variables=drop_variables)

        if 'sst_dtime' in ds.variables and decode_times:
            # replace with new time variable decoding GHRSST time
            dunit = ds.sst_dtime.attrs.get(
                'units', ds.sst_dtime.encoding.get('units'))
            if dunit is None:
                raise ValueError('missing unit for sst_dtime')
            if dunit not in ['second', 'seconds']:
                raise TypeError(
                    'Can not process sst_dtime in {}'.format(dunit)
                )
            reference = np.datetime64(ds['time'].values[0])

            dtime = ds['sst_dtime']

            dtime.attrs = ds['time'].attrs
            dtime.encoding = ds['time'].encoding
            dtime.values = np.ma.array(
                    dtime.values, dtype='timedelta64[s]', copy=False
                ) + reference

            # remove dummy time dimension
            ds = ds.squeeze(dim='time')

            # replace with single time field
            ds = ds.drop_vars(['time'])
            ds = ds.rename({'sst_dtime': 'time'})
            ds = ds.set_coords(['time'])
            if 'comment' in ds['time'].attrs:
                ds['time'].attrs.pop('comment')
            ds['time'].attrs['long_name'] = 'measurement time'

        # normalize collection id (GDS 1.x => 2.x)
        if 'DSD_entry_id' in ds.attrs:
            ds.attrs['id'] = ds.attrs.pop('DSD_entry_id')

        return ds

    open_dataset_parameters = [
        "filename_or_obj", "mask_and_scale", "decode_times",
        "decode_timedelta", "use_cftime", "concat_characters", "decode_coords"]

    def guess_can_open(self, filename_or_obj):
        return re.match(
            r"^[0-9]{8,14}-.*-L[2P|3\w|4]P*_GHRSST-.*nc$",
            Path(filename_or_obj).name) is not None


    description = "Use GHRSST files in Xarray"

    url = "https://link_to/your_backend/documentation"



# -*- coding: utf-8 -*-
"""
Feature classes for time ordered series of grid observation patterns
"""
import datetime
from typing import Tuple, Union

import numpy as np
import xarray as xr

from cerbere.feature.cbasefeature import BaseFeature

from cerbere.feature.cgrid import Grid
from cerbere.feature.ctrajectory import Trajectory


__all__ = ['GridTimeSeries']


class GridIterator:
    """Iterator for looping the grids of the time series"""

    def __init__(self, gridtimeseries):
        self.gridtimeseries = gridtimeseries
        self.index = 0

    def __iter__(self):
        return self

    def next(self):
        if self.index == len(self.gridtimeseries.get_times()):
            raise StopIteration
        self.index = self.index + 1
        return self.gridtimeseries.extract_grid(self.index - 1)


class GridTimeSeries(BaseFeature):
    """Class implementing a time series of grids


    Args:
        content (xarray, mapper or dict of fields): the content from which to
            build the feature coordinates and data fields.

        attrs (dict): a dictionary of global attributes (as in a NetCDF file).
            Make use of the CF attributes. If the feature content is read from
            a file through a mapper, these attributes will be added to the
            existing the list of attributes already in the file (or replace the
            values of the existing attributes with the same name in the file).
            Use with caution in this context.
    """
    cdm_data_type = 'grid'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'longitude': [('X', 'Y'), ('X',)],
        'latitude': [('X', 'Y'), ('Y',)],
        'time': [('T',), ('X', 'Y',)],
        'vertical': ('Z',),
        'data': ('X', 'Y')
    }
    # optional axes for the feature
    cf_optional_axes = ['vertical']


    def __init__(self,
                 dataset,
                 **kwargs
                 ):

        # create feature
        super(GridTimeSeries, self).__init__(
            dataset,
            **kwargs
        )

    def __iter__(self):
        """Returns the iterator"""
        return GridIterator(self)

    @property
    def _grid_class(self):
        return Grid

    def get_geocoord_dimnames(
            self, fieldname: str,
            values: 'xr.DataArray') -> Tuple[str, ...]:
        if fieldname in ['depth', 'height']:
            return 'z',
        elif fieldname == 'time':
            return 'time',
        else:
            return 'y', 'x',

    # def time2t(self, time: 'datetime.datetime'):
    #     """Get the index of the closest time step to the requested time
    #
    #     Args:
    #         time: the time of the closest time step sought in the series
    #     """
    #     numtime = date2num(time, self.get_time_units())
    #     return min(numpy.abs(self.get_times() - numtime).argmin(),
    #                len(self.get_times()) - 1
    #                )

    # def add_grid(self, grid, step=None):
    #     """Add or update a time step to a grid time series.
    #
    #     If the time step already exists, the corresponding values will be
    #     replaced.
    #
    #     Args:
    #         step (int): specify directly the index whithin the time series
    #             where to insert the grid (replacing the current values). If
    #             None, the time of the grid is used to search for the
    #             corresponding index within the time series.
    #     """
    #     time = num2date(grid.get_times()[0], grid.get_time_units())
    #     logging.debug("Source grid time : %s", time)
    #     epsilon = 1
    #     if step is None:
    #         step = 0
    #         steptime = num2date(self.get_times()[step],
    #                             self.get_time_units()
    #                             )
    #         while abs(time - steptime).total_seconds() > epsilon:
    #             step += 1
    #     if step >= len(self.get_times()):
    #         raise Exception('Time not found')
    #     else:
    #         logging.debug('Index in time series : %d', step)
    #         # CASE 1 : time step already exists => update
    #         # check if variable (and corresponding record) already exists
    #         for fieldname in grid.get_fieldnames():
    #             logging.debug('fieldname %s', fieldname)
    #             data = grid.get_values(fieldname)
    #             if self.has_field(fieldname):
    #                 logging.debug('gridtimeserie field already exist')
    #                 logging.debug("ADDING %s %s", fieldname, step)
    #                 field = self.get_field(fieldname)
    #                 logging.debug('field %s', field)
    #                 field._values[step, :, :] = data
    #             else:
    #                 logging.debug(
    #                     '[gridtimeserie] field %s doesnt exist yet in ts', fieldname)
    #                 field_to_add = grid.get_field(fieldname)
    #                 self.add_field(field_to_add)
    #     return



    def grid(self, step: Union[int, 'datetime.datetime', 'np.datetime64']):
        """Extract a grid from a grid time series

        Args:
            step: the grid to extract is defined either by :
                * a time step index (int)
                * a time as a datetime or numpy datetime64

        Return:
            :class:`Grid`: extracted grid.
        """
        if isinstance(step, int):
            grid = self.to_xarray().isel(time=step)
            if 'time' not in grid.time.dims:
                grid.coords['time'] = grid.time.expand_dims(time=1)
            return self._grid_class(grid)

        elif isinstance(step, (datetime.datetime, np.datetime64)):
            step = np.datetime64(step)
            grid = self.to_xarray().sel(time=[step])
            if 'time' not in grid.time.dims:
                grid.coords['time'] = grid.time.expand_dims(time=1)
            return self._grid_class(grid)


    def is_unique_grid_time(self):
        """
        Return True if a unique time is associated with the grid
        (like in L4 products), False if there is time value per pixel
        (like in L3)
        """
        dims = self.get_geolocation_field('time').dimensions.values()[0]
        return (dims == 1)

#
# class CylindricalGridTimeSeries(GridTimeSeries):
#
#     _featureclass_geodimnames = 'time', 'lat', 'lon'
#
#     def __init__(self,
#                  dataset,
#                  **kwargs
#                  ):
#         """
#         Time series of cylindrical lat/lon grids.
#         """
#         # create feature
#         super(CylindricalGridTimeSeries, self).__init__(
#             dataset,
#             **kwargs
#         )
#
#     @property
#     def _grid_class(self):
#         return CylindricalGrid
#
#     def get_geocoord_dimnames(self, fieldname: str, values) -> Tuple[str, ...]:
#         if fieldname in ['depth', 'height']:
#             if len(values.shape) != 1:
#                 raise ValueError('z coordinate must be one-dimensional')
#             return 'z',
#         elif fieldname == 'time':
#             if len(values.shape) != 1:
#                 raise ValueError(
#                     'time coordinate must be one or two dimensional'
#                 )
#             return 'time',
#         else:
#             return 'time', 'lat', 'lon',
#
#     def trajectory(self, trajectory, **kwargs):
#         """
#         Extract the values along the provided trajectory.
#
#         Args:
#             Any args to xarray `:func:interp` function can be used here.
#         """
#         xtraj = trajectory.dataset
#         grid_traj = Trajectory(self.dataset.interp(
#             lat=xtraj.lat,
#             lon=xtraj.lon,
#             time=xtraj.time,
#             method='nearest'
#         ))
#         print(grid_traj)
